/* GDK - The GIMP Drawing Kit
 * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * Modified by the GTK+ Team and others 1997-2000.  See the AUTHORS
 * file for a list of people on the GTK+ Team.  See the ChangeLog
 * files for a list of changes.  These files are distributed with
 * GTK+ at ftp://ftp.gtk.org/pub/gtk/.
 */

#include "config.h"

#include <stdlib.h>
#include <string.h>

#include "gdkcolor.h"
#include "gdkprivate-directfb.h"


#define GDK_COLORMAP_PRIVATE_DATA(cmap) ((GdkColormapDirectFB *) GDK_COLORMAP (cmap)->windowing_data)


static void     gdk_colormap_init        (GdkColormap      *colormap);
static void     gdk_colormap_class_init  (GdkColormapClass *klass);
static void     gdk_colormap_finalize    (GObject          *object);

static gpointer parent_class = NULL;


GType
gdk_colormap_get_type (void)
{
  static GType object_type = 0;

  if (!object_type) {
    static const GTypeInfo object_info =
      {
        sizeof (GdkColormapClass),
        (GBaseInitFunc) NULL,
        (GBaseFinalizeFunc) NULL,
        (GClassInitFunc) gdk_colormap_class_init,
        NULL,           /* class_finalize */
        NULL,           /* class_data */
        sizeof (GdkColormap),
        0,              /* n_preallocs */
        (GInstanceInitFunc) gdk_colormap_init,
      };

    object_type = g_type_register_static (G_TYPE_OBJECT,
                                          "GdkColormap",
                                          &object_info, 0);
  }

  return object_type;
}

static void
gdk_colormap_init (GdkColormap *colormap)
{
  GdkColormapDirectFB *private;

  private = g_new (GdkColormapDirectFB, 1);

  colormap->windowing_data = private;

  colormap->size = 0;
  colormap->colors = NULL;
}

static void
gdk_colormap_class_init (GdkColormapClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);

  parent_class = g_type_class_peek_parent (klass);

  object_class->finalize = gdk_colormap_finalize;
}

static void
gdk_colormap_finalize (GObject *object)
{
  GdkColormap *colormap = GDK_COLORMAP (object);
//     GdkColormapPrivateDirectFB *private = GDK_COLORMAP_PRIVATE_DATA (colormap);

  g_free (colormap->colors);

  G_OBJECT_CLASS (parent_class)->finalize (object);
}

GdkColormap*
gdk_colormap_new (GdkVisual *visual,
                  gboolean   private_cmap)
{
  GdkColormap *colormap;
//     GdkColormapPrivateDirectFB *private;

  g_return_val_if_fail (visual != NULL, NULL);

  colormap = g_object_new (gdk_colormap_get_type (), NULL);
//     private = GDK_COLORMAP_PRIVATE_DATA (colormap);

  colormap->visual = visual;

  colormap->size = visual->colormap_size;

  return colormap;
}

void
gdk_colormap_sync (GdkColormap *colormap,
           gboolean     force)
{
  g_message("unimplemented %s", __FUNCTION__);
}

GdkColormap*
gdk_colormap_get_system (void)
{
  return gdk_colormap_new(gdk_visual_get_system(), 0);
}


gint
gdk_colormap_get_system_size (void)
{
  g_message("unimplemented %s", __FUNCTION__);

  return 65536;
}

void
gdk_colormap_change (GdkColormap *colormap,
             gint         ncolors)
{
  g_message("unimplemented %s", __FUNCTION__);
}

gboolean
gdk_colors_alloc (GdkColormap   *colormap,
                  gboolean       contiguous,
                  gulong        *planes,
                  gint           nplanes,
                  gulong        *pixels,
                  gint           npixels)
{
  g_message("unimplemented %s", __FUNCTION__);
   
  return FALSE;
}

static struct cspec 
{
  gchar *name;
  guint  red, green, blue;
} cnames [] = {
  { "white",   0xffff, 0xffff, 0xffff },
  { "black",   0,      0,      0      },
  { "red",     0xffff, 0,      0      },
  { "green",   0,      0xffff, 0      },
  { "blue",    0,      0,      0xffff },
  { "cyan",    0,      0xffff, 0xffff },
  { "magenta", 0xffff, 0xffff, 0      },
  { "yellow",  0,      0xffff, 0xffff },
  { NULL }
};

gboolean
gdk_color_parse (const gchar *spec,
                 GdkColor    *color)
{
  guint   red, green, blue;
  guint   csize;
  guint   i;
  gdouble dval;
  gchar  *end;
  guint   scale[] = { 0, 4096, 256, 16, 1 };
  guint   add[]   = { 0, 4095, 255, 15, 1 };

  g_return_val_if_fail (spec != NULL, FALSE);
  g_return_val_if_fail (color != NULL, FALSE);

  /*  to be removed  */
  g_message ("color parsing %s", spec);

  if (*spec == '#')
    {
      guint len;
      guint result;

      len = strlen (spec) - 1;
      switch (len)
        {
        case 3:
          result = sscanf (spec + 1, "%1x%1x%1x", &red, &green, &blue);
          break;
        case 6:
          result = sscanf (spec + 1, "%2x%2x%2x", &red, &green, &blue);
          break;
        case 9:
          result = sscanf (spec + 1, "%3x%3x%3x", &red, &green, &blue);
          break;
        case 12:
          result = sscanf (spec + 1, "%4x%4x%4x", &red, &green, &blue);
          break;
        default:
          result = 0;
          break;
        }
      
      if (result != 3)
        return FALSE;
          
      switch (len)
        {
        case 3:
          color->red   = red   << 12;
          color->green = green << 12;
          color->blue  = blue  << 12;
          break;
        case 6:
          color->red   = red   << 8;
          color->green = green << 8;
          color->blue  = blue  << 8;
          break;
        case 9:
          color->red   = red   << 4;
          color->green = green << 4;
          color->blue  = blue  << 4;
          break;
        case 12:
          color->red   = red;
          color->green = green;
          color->blue  = blue;
          break;
        default:
          break;
        }

      return TRUE;
    }

  if (strncmp (spec, "rgb:", 4) == 0) 
    {
      spec += 4;

      /* red */
      color->red = strtol (spec, &end, 16);
      if (end == spec || *end != '/')
        return FALSE;
      csize = end-spec;
      color->red *= scale[csize];
      color->red += add[csize];
      spec += csize + 1;

      /* green */
      color->green = strtol(spec, &end, 16);
      if (end == spec || *end != '/')
        return FALSE;
      csize = end-spec;
      color->green *= scale[csize];
      color->green += add[csize];
      spec += csize + 1;

      /* blue */
      color->blue = strtol(spec, &end, 16);
      if (end == spec || *end != '\0')
        return FALSE;
      csize = end-spec;
      color->blue *= scale[csize];
      color->blue += add[csize];

      return TRUE;
    } 

  if (strncmp (spec, "rgbi:", 5) == 0) 
    {
      spec += 5;

      /* red */
      dval = strtod (spec, &end);
      if (end == spec || *end != '/' || dval > 1.0 || dval < 0)
        return FALSE;
      color->red = dval * 0xffff;
      spec += end-spec + 1;

      /* green */
      dval = strtod(spec, &end);
      if (end == spec || *end != '/' || dval > 1.0 || dval < 0)
        return FALSE;
      color->green = dval * 0xffff;
      spec += end-spec + 1;

      /* blue */
      dval = strtod(spec, &end);
      if (end == spec || *end != '0' || dval > 1.0 || dval < 0)
        return 0;
      color->blue = dval*0xffff;

      return TRUE;
    } 

  /* use a cdb database, instead, later */
  for (i = 0; cnames[i].name; ++i) 
    {
      if (strcmp (cnames[i].name, spec))
        continue;
      
      color->red   = cnames[i].red;
      color->green = cnames[i].green;
      color->blue  = cnames[i].blue;
      return TRUE;
    }
  
  if (spec[0] == 'g' && spec[1] == 'r' && 
      (spec[2] == 'a' || spec[2] == 'e') && spec[3] == 'y') 
    {
      dval = strtol (spec + 4, NULL, 10) / 100;
      color->red = color->green = color->blue = 255 * dval;
      return TRUE;
    }

  return FALSE;
}

void
gdk_colors_free (GdkColormap *colormap,
                 gulong      *in_pixels,
                 gint         in_npixels,
                 gulong       planes)
{
  g_message("unimplemented %s", __FUNCTION__);
}

void
gdk_colormap_free_colors (GdkColormap *colormap,
                          GdkColor    *colors,
                          gint         ncolors)
{
}

gint
gdk_colormap_alloc_colors (GdkColormap *colormap,
                           GdkColor    *colors,
                           gint         ncolors,
                           gboolean     writeable,
                           gboolean     best_match,
                           gboolean    *success)
{
  GdkColormapDirectFB *private;
  GdkVisual           *visual;
  gint i;

  g_return_val_if_fail (colormap != NULL, FALSE);
  g_return_val_if_fail (GDK_IS_COLORMAP (colormap), FALSE);

  g_return_val_if_fail (colors != NULL, FALSE);

  private = GDK_COLORMAP_PRIVATE_DATA (colormap);

  for (i = 0; i < ncolors; i++)
    {
      success[i] = FALSE;
    }
  
  switch (colormap->visual->type)
    {
    case GDK_VISUAL_PSEUDO_COLOR:
    case GDK_VISUAL_GRAYSCALE:
      //TODODO
      break;
      
    case GDK_VISUAL_TRUE_COLOR:
      visual = colormap->visual;
      
      for (i = 0; i < ncolors; i++)
        {
          colors[i].pixel = (((colors[i].red  >> (16 - visual->red_prec)) << visual->red_shift) +
                             ((colors[i].green >> (16 - visual->green_prec)) << visual->green_shift) +
                             ((colors[i].blue  >> (16 - visual->blue_prec)) << visual->blue_shift));
          success[i] = TRUE;
        }
      break;
      
    case GDK_VISUAL_STATIC_GRAY:
    case GDK_VISUAL_STATIC_COLOR:
      //TODODO
      break;
      
    case GDK_VISUAL_DIRECT_COLOR:
      g_assert_not_reached ();
    }

  return 0;
}

gboolean
gdk_color_change (GdkColormap *colormap,
                  GdkColor    *color)
{
  g_message("unimplemented %s", __FUNCTION__);

  return TRUE;
}

void
gdk_colormap_query_color (GdkColormap *colormap,
                          gulong       pixel,
                          GdkColor    *result)
{
  GdkVisual *visual;

  g_return_if_fail (GDK_IS_COLORMAP (colormap));
  
  visual = gdk_colormap_get_visual (colormap);

  switch (visual->type) 
    {
    case GDK_VISUAL_DIRECT_COLOR:
    case GDK_VISUAL_TRUE_COLOR:
      result->red = 65535. *
                   (gdouble)((pixel & visual->red_mask) >> visual->red_shift) /
                   ((1 << visual->red_prec) - 1);
      
      result->green = 65535. *
                     (gdouble)((pixel & visual->green_mask) >> visual->green_shift) /
                     ((1 << visual->green_prec) - 1);
      
      result->blue = 65535. *
                    (gdouble)((pixel & visual->blue_mask) >> visual->blue_shift) /
                    ((1 << visual->blue_prec) - 1);
      break;
      
    case GDK_VISUAL_STATIC_GRAY:
    case GDK_VISUAL_GRAYSCALE:
      result->red   =
      result->green =
      result->blue  = 65535. * (double)pixel/((1<<visual->depth) - 1);
      break;
      
    case GDK_VISUAL_PSEUDO_COLOR:
      result->red   = colormap->colors[pixel].red;
      result->green = colormap->colors[pixel].green;
      result->blue  = colormap->colors[pixel].blue;
      break;
      
    default:
      g_assert_not_reached ();
      break;
    }
}

