/* GDK - The GIMP Drawing Kit
 * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * Modified by the GTK+ Team and others 1997-2000.  See the AUTHORS
 * file for a list of people on the GTK+ Team.  See the ChangeLog
 * files for a list of changes.  These files are distributed with
 * GTK+ at ftp://ftp.gtk.org/pub/gtk/.
 */

#include <stdlib.h>

#include <pango/pangoft2.h>
#include <freetype/ftglyph.h>

#include "gdkinternals.h"
#include "gdkprivate-directfb.h"

#include "glyphsurfacecache.h"

#include "gdkregion-generic.h"


static void gdk_drawable_impl_directfb_class_init (GdkDrawableImplDirectFBClass *klass);
static void gdk_directfb_draw_lines   (GdkDrawable *drawable,
                                       GdkGC       *gc,
                                       GdkPoint    *points,
                                       gint         npoints);


static gpointer  parent_class = NULL;


/**********************************************************
 * DirectFB specific implementations of generic functions *
 **********************************************************/

static GdkColormap*
gdk_directfb_get_colormap (GdkDrawable *drawable)
{
  GdkColormap *retval;

  retval = GDK_DRAWABLE_IMPL_DIRECTFB (drawable)->colormap;

  if (!retval)
    retval = gdk_colormap_get_system ();

  return retval;
}

static void
gdk_directfb_set_colormap (GdkDrawable *drawable,
                           GdkColormap *colormap)
{
  GdkDrawableImplDirectFB *impl;

  impl = GDK_DRAWABLE_IMPL_DIRECTFB (drawable);

  if (impl->colormap == colormap)
    return;

  if (impl->colormap)
    gdk_colormap_unref (impl->colormap);

  impl->colormap = colormap;

  if (impl->colormap)
    gdk_colormap_ref (impl->colormap);
}

static gint
gdk_directfb_get_depth (GdkDrawable *drawable)
{
  GdkDrawableImplDirectFB *impl;

  impl = GDK_DRAWABLE_IMPL_DIRECTFB (drawable);

  return BITS_PER_PIXEL(impl->format);
}

static void
gdk_directfb_get_size (GdkDrawable *drawable,
                       gint        *width,
                       gint        *height)
{
  GdkDrawableImplDirectFB *impl;

  impl = GDK_DRAWABLE_IMPL_DIRECTFB (drawable);

  if (width)
    *width = impl->width;

  if (height)
    *height = impl->height;
}

static GdkVisual*
gdk_directfb_get_visual (GdkDrawable *drawable)
{
  GdkDrawableImplDirectFB *impl;

  impl = GDK_DRAWABLE_IMPL_DIRECTFB (drawable);

  return gdk_drawable_get_visual (impl->wrapper);
}

/* Calculates the real clipping region for a drawable, taking into account
 * other windows, gc clip region and gc clip mask.
 */
GdkRegion *
gdk_directfb_clip_region (GdkDrawable *drawable,
                          GdkGC       *gc,
                          gboolean     do_clipping,
                          gboolean     do_children,
                          gboolean     full_shapes)
{
  GdkRectangle             draw_rect;
  GdkRegion               *clip_region;
  GdkRegion               *tmpreg;
  gboolean                 skipit = FALSE;
  GdkDrawableImplDirectFB *private;

  g_return_val_if_fail (drawable != NULL, NULL);
  g_return_val_if_fail (GDK_IS_DRAWABLE (drawable), NULL);

  g_return_val_if_fail (GDK_IS_DRAWABLE_IMPL_DIRECTFB (drawable), NULL);

  private = GDK_DRAWABLE_IMPL_DIRECTFB (drawable);

  g_assert (!GDK_IS_WINDOW (private->wrapper) ||
            !GDK_WINDOW_OBJECT (private->wrapper)->input_only);

  draw_rect.x = 0;
  draw_rect.y = 0;

  if (!GDK_IS_WINDOW (private) || GDK_WINDOW_IS_MAPPED (private))
    {
      draw_rect.width  = private->width;
      draw_rect.height = private->height;
    }
  else
    {
      draw_rect.width  = 0;
      draw_rect.height = 0;
      skipit = TRUE;
    }
  clip_region = gdk_region_rectangle (&draw_rect);

  if (skipit)
    return clip_region;

  if (private->buffered && private->paint_region)
    gdk_region_intersect (clip_region, private->paint_region); 

  if (gc && 
      GDK_GC_DIRECTFB (gc)->values.subwindow_mode == GDK_INCLUDE_INFERIORS)
    do_children = FALSE;

  if (do_clipping &&
      GDK_IS_WINDOW (private->wrapper) &&
      GDK_WINDOW_IS_MAPPED (private->wrapper) &&
      !GDK_WINDOW_OBJECT (private->wrapper)->input_only)
    {
      if (do_children)
        {
          GList *cur;

          for (cur = GDK_WINDOW_OBJECT (private->wrapper)->children;
               cur;
               cur = cur->next)
            {
              GdkWindowObject         *cur_private;
              GdkDrawableImplDirectFB *cur_impl;

              cur_private = GDK_WINDOW_OBJECT (cur->data);

              if (!GDK_WINDOW_IS_MAPPED (cur_private) || 
                  cur_private->input_only)
                continue;

              cur_impl = GDK_DRAWABLE_IMPL_DIRECTFB (cur_private->impl);
              draw_rect.x      = cur_private->x;
              draw_rect.y      = cur_private->y;
              draw_rect.width  = cur_impl->width;
              draw_rect.height = cur_impl->height;

              /*g_message("clipping child at: %d, %d with size %d*%d",
                draw_rect.x, draw_rect.y,
                draw_rect.width, draw_rect.height );*/
              
              tmpreg = gdk_region_rectangle (&draw_rect);
              
              gdk_region_subtract (clip_region, tmpreg);
              gdk_region_destroy (tmpreg);
            }
        }
    }
  
  if (gc)
    {
      if (GDK_GC_DIRECTFB (gc)->clip_region)
        {
          tmpreg = gdk_region_copy (GDK_GC_DIRECTFB (gc)->clip_region);
          
          gdk_region_offset (tmpreg, GDK_GC (gc)->clip_x_origin,
                             GDK_GC (gc)->clip_y_origin);
          gdk_region_intersect (clip_region, tmpreg);
          gdk_region_destroy (tmpreg);
        }
      
    }

  return clip_region;
}

/* Drawing
 */

void
gdk_directfb_draw_rectangle (GdkDrawable *drawable,
                             GdkGC       *gc,
                             gint         filled,
                             gint         x,
                             gint         y,
                             gint         width,
                             gint         height)
{
  GdkDrawableImplDirectFB *impl;
  GdkGCDirectFB           *gc_private;
  GdkRectangle  tmprect;
  GdkRegion    *tmpreg;
  GdkRegion    *clip_region;
  gint          i;

  g_return_if_fail (drawable != NULL);
  g_return_if_fail (GDK_IS_DRAWABLE (drawable));

  impl = GDK_DRAWABLE_IMPL_DIRECTFB (drawable);

  if (!impl->surface)
    return;

  if (gc)
    {
      gc_private = GDK_GC_DIRECTFB (gc);

      /*  FIXME: quick hack for Zappa  */
      if (gc_private->values.foreground.red   <= 0x00FF &&
	  gc_private->values.foreground.green <= 0x00FF &&
	  gc_private->values.foreground.blue  <= 0x00FF)
	{
	  impl->surface->SetColor( impl->surface, 0, 0, 0, 0);
	}
      else
        {
	  impl->surface->SetColor( impl->surface,
				   gc_private->values.foreground.red   >> 8,
				   gc_private->values.foreground.green >> 8,
				   gc_private->values.foreground.blue  >> 8,
				   0xFF );
	}
    }
  else
    {
      GdkWindowObject *win = GDK_WINDOW_OBJECT (impl->wrapper);

      impl->surface->SetColor( impl->surface,
                               win->bg_color.red   >> 8,
                               win->bg_color.green >> 8,
                               win->bg_color.blue  >> 8,
                               0x00 );
    }

  if (filled)
    {
      clip_region = gdk_directfb_clip_region (drawable, gc,
                                              TRUE, TRUE, TRUE);

      tmprect.x = x;
      tmprect.y = y;
      tmprect.width  = width;
      tmprect.height = height;
      tmpreg = gdk_region_rectangle (&tmprect);

      gdk_region_intersect (tmpreg, clip_region);

      for (i = 0; i < tmpreg->numRects; i++)
        {
          impl->surface->FillRectangle( impl->surface,
                                        tmpreg->rects[i].x1,
                                        tmpreg->rects[i].y1,
                                        tmpreg->rects[i].x2 - tmpreg->rects[i].x1,
                                        tmpreg->rects[i].y2 - tmpreg->rects[i].y1);

          gdk_directfb_add_update (impl, (DFBRegion *) &tmpreg->rects[i]);
        }

      gdk_region_destroy (tmpreg);
      gdk_region_destroy (clip_region);
    }
  else
    {
      DFBRegion region = { x, y, x + width, y + height };
      
      /*  DirectFB does not draw rectangles the X way. Using DirectFB,
          a filled Rectangle has the same size as a drawn one, while
          X draws the rectangle one pixel taller and wider.  */
      impl->surface->DrawRectangle (impl->surface,
                                    x, y, width + 1, height + 1);
      
      gdk_directfb_add_update (impl, &region);
    }
}

static void
gdk_directfb_draw_arc (GdkDrawable *drawable,
                       GdkGC       *gc,
                       gint         filled,
                       gint         x,
                       gint         y,
                       gint         width,
                       gint         height,
                       gint         angle1,
                       gint         angle2)
{
  g_message("unimplemented %s", __FUNCTION__);
}

static void
gdk_directfb_draw_polygon (GdkDrawable *drawable,
                           GdkGC       *gc,
                           gint         filled,
                           GdkPoint    *points,
                           gint         npoints)
{
  if (filled)
    g_message("filled polygons are not yet supported, drawing outlines");

  if (points[0].x != points[npoints-1].x ||
      points[0].y != points[npoints-1].y)
    {
      GdkPoint *tmp_points;

      tmp_points = g_new (GdkPoint, npoints + 1);
      memcpy (tmp_points, points, npoints * sizeof (GdkPoint));
      tmp_points[npoints].x = points[0].x;
      tmp_points[npoints].y = points[0].y;

      gdk_directfb_draw_lines (drawable, gc, tmp_points, npoints + 1);

      g_free (tmp_points);
    }
  else
    {
      gdk_directfb_draw_lines (drawable, gc, points, npoints);
    }
}

static void
gdk_directfb_draw_text (GdkDrawable *drawable,
                        GdkFont     *font,
                        GdkGC       *gc,
                        gint         x,
                        gint         y,
                        const gchar *text,
                        gint         text_length)
{
  g_message("unimplemented %s", __FUNCTION__);
}

static void
gdk_directfb_draw_text_wc (GdkDrawable    *drawable,
                           GdkFont        *font,
                           GdkGC          *gc,
                           gint            x,
                           gint            y,
                           const GdkWChar *text,
                           gint            text_length)
{
  g_message("unimplemented %s", __FUNCTION__);
}

static void
gdk_directfb_draw_drawable (GdkDrawable *drawable,
                            GdkGC       *gc,
                            GdkDrawable *src,
                            gint         xsrc,
                            gint         ysrc,
                            gint         xdest,
                            gint         ydest,
                            gint         width,
                            gint         height)
{
  GdkDrawableImplDirectFB *impl;
  GdkDrawableImplDirectFB *src_impl;

  DFBRectangle rect = { xsrc, ysrc, width, height };
  DFBRegion region  = { xdest, ydest, xdest + width - 1, ydest + height - 1 };

  impl = GDK_DRAWABLE_IMPL_DIRECTFB (drawable);

  if (!impl->surface)
    return;

  if (GDK_IS_PIXMAP (src))
    src_impl = GDK_DRAWABLE_IMPL_DIRECTFB (GDK_PIXMAP_OBJECT (src)->impl);
  else if (GDK_IS_WINDOW (src))
    src_impl = GDK_DRAWABLE_IMPL_DIRECTFB (GDK_WINDOW_OBJECT (src)->impl);
  else if (GDK_IS_DRAWABLE_IMPL_DIRECTFB (src))
    src_impl = GDK_DRAWABLE_IMPL_DIRECTFB (src);
  else
    return;

  impl->surface->Blit( impl->surface, src_impl->surface,
                       &rect, xdest, ydest );

  gdk_directfb_add_update (impl, &region);
}

static void
gdk_directfb_draw_points (GdkDrawable *drawable,
                          GdkGC       *gc,
                          GdkPoint    *points,
                          gint         npoints)
{
  g_message("unimplemented %s", __FUNCTION__);
}

static void
gdk_directfb_draw_segments (GdkDrawable *drawable,
                            GdkGC       *gc,
                            GdkSegment  *segs,
                            gint         nsegs)
{
  GdkDrawableImplDirectFB *impl;
  GdkGCDirectFB           *gc_private = GDK_GC_DIRECTFB (gc);

  DFBRegion region = { segs->x1, segs->y1, segs->x2, segs->y2 };

  impl = GDK_DRAWABLE_IMPL_DIRECTFB (drawable);

  if (!impl->surface)
    return;

  impl->surface->SetColor( impl->surface,
                           gc_private->values.foreground.red   >> 8,
                           gc_private->values.foreground.green >> 8,
                           gc_private->values.foreground.blue  >> 8,
                           0xFF );

  while (nsegs > 0)
    {
      impl->surface->DrawLine( impl->surface,
                               segs->x1, segs->y1, segs->x2, segs->y2 );

      if (segs->x1 < region.x1)
        region.x1 = segs->x1;
      if (segs->x2 < region.x1)
        region.x1 = segs->x2;

      if (segs->y1 < region.y1)
        region.y1 = segs->y1;
      if (segs->y2 < region.y1)
        region.y1 = segs->y2;

      if (segs->x1 > region.x2)
        region.x2 = segs->x1;
      if (segs->x2 > region.x2)
        region.x2 = segs->x2;

      if (segs->y1 > region.y2)
        region.y2 = segs->y1;
      if (segs->y2 > region.y2)
        region.y2 = segs->y2;

/*      g_message("SEGMENT: %d, %d to %d, %d - %x : %x %x %x",
                segs->x1, segs->y1, segs->x2, segs->y2,
                gc_private->values.foreground.pixel,
                gc_private->values.foreground.red,
                gc_private->values.foreground.green,
                gc_private->values.foreground.blue);*/

      nsegs--;
      segs++;
    }

  gdk_directfb_add_update (impl, &region);
}

static void
gdk_directfb_draw_lines (GdkDrawable *drawable,
                         GdkGC       *gc,
                         GdkPoint    *points,
                         gint         npoints)
{
  GdkDrawableImplDirectFB *impl;
  GdkGCDirectFB           *gc_private = GDK_GC_DIRECTFB (gc);

  DFBRegion region = { points->x, points->y, points->x, points->y };

  impl = GDK_DRAWABLE_IMPL_DIRECTFB (drawable);

  if (!impl->surface)
    return;

  impl->surface->SetColor( impl->surface,
                           gc_private->values.foreground.red   >> 8,
                           gc_private->values.foreground.green >> 8,
                           gc_private->values.foreground.blue  >> 8,
                           0xFF );

  while (npoints-- > 1) {
    impl->surface->DrawLine( impl->surface,
                             points[0].x, points[0].y,
                             points[1].x, points[1].y );

    if (points->x < region.x1)
      region.x1 = points->x;

    if (points->y < region.y1)
      region.y1 = points->y;

    if (points->x > region.x2)
      region.x2 = points->x;

    if (points->y > region.y2)
      region.y2 = points->y;

    points++;
  }

  gdk_directfb_add_update (impl, &region);
}

static void
gdk_directfb_draw_glyphs (GdkDrawable      *drawable,
                          GdkGC            *gc,
                          PangoFont        *font,
                          gint              x,
                          gint              y,
                          PangoGlyphString *glyphs)
{
  IDirectFBSurface      *surface;

  PangoGlyphInfo          *gi;

  DFBRectangle             rect;
  GdkRectangle             dest_rect;
  GdkDrawableImplDirectFB *impl;
  GdkGCDirectFB           *gc_private;
  GdkRegion               *tmpreg;
  GdkRegion               *clip_region;
  gint                     i, j;
  gint                     xpos;
  gint                     top, left;

  __u8  r, g, b;

  g_return_if_fail (drawable != NULL);
  g_return_if_fail (gc != NULL);
  g_return_if_fail (font != NULL);
  g_return_if_fail (glyphs != NULL);

  impl = GDK_DRAWABLE_IMPL_DIRECTFB (drawable);
  gc_private = GDK_GC_DIRECTFB (gc);

  if (!impl->surface)
    return;

  clip_region = gdk_directfb_clip_region (drawable, gc, TRUE, TRUE, TRUE);

  r = gc_private->values.foreground.red   >> 8;
  g = gc_private->values.foreground.green >> 8;
  b = gc_private->values.foreground.blue  >> 8;
  impl->surface->SetColor (impl->surface, r, g, b, 0xFF);
  
  impl->surface->SetBlittingFlags (impl->surface, (DSBLIT_BLEND_ALPHACHANNEL |
                                                   DSBLIT_COLORIZE));

  xpos = 0;

  for (gi = glyphs->glyphs, i = 0; 
       i < glyphs->num_glyphs; 
       gi++, i++)
    {
      if (gi->glyph &&
          glyph_surface_cache_get_surface (glyph_surface_cache, 
                                           font, gi->glyph,
                                           &surface, &rect, &top, &left))
        {
          dest_rect.x = x + PANGO_PIXELS (xpos + gi->geometry.x_offset) + left;
          dest_rect.y = y + PANGO_PIXELS (gi->geometry.y_offset) - 1 - top;
          dest_rect.width  = rect.w;
          dest_rect.height = rect.h;

          tmpreg = gdk_region_rectangle (&dest_rect);
          gdk_region_intersect (tmpreg, clip_region);
          
          for (j = 0; j < tmpreg->numRects; j++)
            {
              impl->surface->SetClip (impl->surface, 
                                      (DFBRegion *) &tmpreg->rects[j]);
              impl->surface->Blit (impl->surface, surface, &rect, 
                                   dest_rect.x, dest_rect.y);
              
              gdk_directfb_add_update (impl, (DFBRegion *) &tmpreg->rects[j]);
            }
          
          gdk_region_destroy (tmpreg); 
      
          surface->Release (surface);
          if (surface->priv == NULL)
            free (surface);
        } 

      xpos += gi->geometry.width;
    }
  
  gdk_region_destroy (clip_region);

  impl->surface->SetClip (impl->surface, NULL);
  impl->surface->SetBlittingFlags (impl->surface, DSBLIT_NOFX);
 
#ifdef CACHE_INFO
  {
    gint size, hits, misses;
    glyph_surface_cache_get_info (glyph_surface_cache, &size, &hits, &misses);
    g_print ("GlyphSurfaceCache (%d) %4.1f %% hits\n", 
             size, 100.0 * hits / (gdouble) (hits + misses));
  } 
#endif
}

static void
gdk_directfb_draw_image (GdkDrawable *drawable,
                         GdkGC       *gc,
                         GdkImage    *image,
                         gint         xsrc,
                         gint         ysrc,
                         gint         xdest,
                         gint         ydest,
                         gint         width,
                         gint         height)
{
  GdkDrawableImplDirectFB *impl;
  GdkImageDirectFB        *image_private;

  DFBRectangle rect   = { xsrc, ysrc, width, height };
  DFBRegion    region = { xdest, ydest, xdest + width - 1, ydest + height - 1 };

  gint pitch = 0;

  impl = GDK_DRAWABLE_IMPL_DIRECTFB (drawable);
  image_private = image->windowing_data;

  if (!impl->surface)
    return;

  image_private->surface->Unlock( image_private->surface );

  impl->surface->Blit( impl->surface, image_private->surface,
                       &rect, xdest, ydest );

  gdk_directfb_add_update( impl, &region );

  image_private->surface->Lock( image_private->surface, DSLF_WRITE,
                                &image->mem, &pitch );

  image->bpl = pitch;
}

/*
 * Object stuff
 */

static void
gdk_drawable_impl_directfb_class_init (GdkDrawableImplDirectFBClass *klass)
{
  GdkDrawableClass *drawable_class = GDK_DRAWABLE_CLASS (klass);

  parent_class = g_type_class_peek_parent (klass);

  drawable_class->create_gc      = _gdk_directfb_gc_new;
  drawable_class->draw_rectangle = gdk_directfb_draw_rectangle;
  drawable_class->draw_arc       = gdk_directfb_draw_arc;
  drawable_class->draw_polygon   = gdk_directfb_draw_polygon;
  drawable_class->draw_text      = gdk_directfb_draw_text;
  drawable_class->draw_text_wc   = gdk_directfb_draw_text_wc;
  drawable_class->draw_drawable  = gdk_directfb_draw_drawable;
  drawable_class->draw_points    = gdk_directfb_draw_points;
  drawable_class->draw_segments  = gdk_directfb_draw_segments;
  drawable_class->draw_lines     = gdk_directfb_draw_lines;
  drawable_class->draw_glyphs    = gdk_directfb_draw_glyphs;
  drawable_class->draw_image     = gdk_directfb_draw_image;

  drawable_class->set_colormap   = gdk_directfb_set_colormap;
  drawable_class->get_colormap   = gdk_directfb_get_colormap;

  drawable_class->get_depth      = gdk_directfb_get_depth;
  drawable_class->get_visual     = gdk_directfb_get_visual;

  drawable_class->get_image      = _gdk_directfb_get_image;
  drawable_class->get_size       = gdk_directfb_get_size;
}

GType
gdk_drawable_impl_directfb_get_type (void)
{
  static GType object_type = 0;

  if (!object_type)
    {
      static const GTypeInfo object_info =
        {
          sizeof (GdkDrawableImplDirectFBClass),
          (GBaseInitFunc) NULL,
          (GBaseFinalizeFunc) NULL,
          (GClassInitFunc) gdk_drawable_impl_directfb_class_init,
          NULL,           /* class_finalize */
          NULL,           /* class_data */
          sizeof (GdkDrawableImplDirectFB),
          0,              /* n_preallocs */
          (GInstanceInitFunc) NULL,
        };

      object_type = g_type_register_static (GDK_TYPE_DRAWABLE,
                                            "GdkDrawableImplDirectFB",
                                            &object_info, 0);
    }

  return object_type;
}

void
gdk_directfb_add_update (GdkDrawableImplDirectFB *impl,
                         DFBRegion               *region)
{
  g_return_if_fail (impl != NULL);
  g_return_if_fail (GDK_IS_DRAWABLE_IMPL_DIRECTFB (impl));

  if (!impl->surface)
    return;

  if (!impl->buffered)
    {
      if (impl->update_region_valid)
        {
          if (region)
            {
              if (region->x1 < impl->update_region.x1)
                impl->update_region.x1 = region->x1;
              if (region->y1 < impl->update_region.y1)
                impl->update_region.y1 = region->y1;
              if (region->x2 > impl->update_region.x2)
                impl->update_region.x2 = region->x2;
              if (region->y2 > impl->update_region.y2)
                impl->update_region.y2 = region->y2;
            }
          
          impl->surface->Flip( impl->surface, &impl->update_region, 0 );
          impl->update_region_valid = FALSE;
        }
      else if (region)
        {
          impl->surface->Flip( impl->surface, region, 0 );
        }
    }
  else if (region)
    {
      if (impl->update_region_valid)
        {
          if (region->x1 < impl->update_region.x1)
            impl->update_region.x1 = region->x1;
          if (region->y1 < impl->update_region.y1)
            impl->update_region.y1 = region->y1;
          if (region->x2 > impl->update_region.x2)
            impl->update_region.x2 = region->x2;
          if (region->y2 > impl->update_region.y2)
            impl->update_region.y2 = region->y2;
        }
      else
        {
          impl->update_region       = *region;
          impl->update_region_valid = TRUE;
        }
    }
}
