/* GDK - The GIMP Drawing Kit
 * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * Modified by the GTK+ Team and others 1997-2000.  See the AUTHORS
 * file for a list of people on the GTK+ Team.  See the ChangeLog
 * files for a list of changes.  These files are distributed with
 * GTK+ at ftp://ftp.gtk.org/pub/gtk/.
 */

#include "config.h"

#include <stdlib.h>

#include "gdkimage.h"
#include "gdkinternals.h"
#include "gdkprivate-directfb.h"


static GList    *image_list   = NULL;
static gpointer  parent_class = NULL;

static void gdk_directfb_image_destroy (GdkImage      *image);
static void gdk_image_init             (GdkImage      *image);
static void gdk_image_class_init       (GdkImageClass *klass);
static void gdk_image_finalize         (GObject       *object);

GType
gdk_image_get_type (void)
{
  static GType object_type = 0;

  if (!object_type)
    {
      static const GTypeInfo object_info =
        {
          sizeof (GdkImageClass),
          (GBaseInitFunc) NULL,
          (GBaseFinalizeFunc) NULL,
          (GClassInitFunc) gdk_image_class_init,
          NULL,           /* class_finalize */
          NULL,           /* class_data */
          sizeof (GdkImage),
          0,              /* n_preallocs */
          (GInstanceInitFunc) gdk_image_init,
        };

      object_type = g_type_register_static (G_TYPE_OBJECT,
                                            "GdkImage",
                                            &object_info, 0);
    }

  return object_type;
}

static void
gdk_image_init (GdkImage *image)
{
  image->windowing_data = g_new0 (GdkImageDirectFB, 1);
  image->mem = NULL;

  image_list = g_list_prepend (image_list, image);
}

static void
gdk_image_class_init (GdkImageClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);

  parent_class = g_type_class_peek_parent (klass);

  object_class->finalize = gdk_image_finalize;
}

static void
gdk_image_finalize (GObject *object)
{
  GdkImage *image;

  image = GDK_IMAGE (object);

  if (image->depth == 1)
    g_free (image->mem);

  gdk_directfb_image_destroy (image);

  image_list = g_list_remove (image_list, image);

  if (G_OBJECT_CLASS (parent_class)->finalize)
    G_OBJECT_CLASS (parent_class)->finalize (object);
}

void
gdk_image_exit (void)
{
  GObject *image;

  while (image_list)
    {
      image = image_list->data;
      gdk_image_finalize (image);
    }
}

GdkImage *
gdk_image_new_bitmap (GdkVisual *visual,
                      gpointer   data,
                      gint       w,
                      gint       h)
{
  GdkImage         *image;
  GdkImageDirectFB *private;

  image = g_object_new (gdk_image_get_type (), NULL);
  private = image->windowing_data;

  image->type   = GDK_IMAGE_SHARED;
  image->visual = visual;
  image->width  = w;
  image->height = h;
  image->depth  = 1;

  GDK_NOTE (MISC, g_print ("gdk_image_new_bitmap: %dx%d\n", w, h));

  g_message("not fully implemented %s", __FUNCTION__);

  image->bpl = (w + 7) / 8;
  image->mem = g_malloc (image->bpl * h);
  image->byte_order = GDK_LSB_FIRST;

  image->bpp = 1;

  return image;
}

void
_gdk_windowing_image_init (void)
{
}

GdkImage*
gdk_image_new (GdkImageType  type,
               GdkVisual    *visual,
               gint          width,
               gint          height)
{
  GdkImage *image;
  GdkImageDirectFB *private;
  DFBResult ret;
  int pitch;
  DFBSurfacePixelFormat format;
  IDirectFBSurface *surface;
  DFBSurfaceDescription desc;

  if (type == GDK_IMAGE_FASTEST || type == GDK_IMAGE_NORMAL)
    type = GDK_IMAGE_SHARED;

  GDK_NOTE (MISC, g_print ("gdk_image_new: %dx%dx%d %s\n",
                           width, height, visual->depth,
                           (type == GDK_IMAGE_SHARED ? "shared" :
                            "???")));

  desc.flags  = DSDESC_BPP | DSDESC_WIDTH | DSDESC_HEIGHT;
  desc.width  = width;
  desc.height = height;
  desc.bpp    = visual->depth;

  ret = DirectFB->CreateSurface( DirectFB, &desc, &surface );
  if (ret)
    {
      DirectFBError( "gdk_image_new - CreateSurface", ret );
      return NULL;
    }
  surface->GetPixelFormat( surface, &format );

  image = g_object_new (gdk_image_get_type (), NULL);
  private = image->windowing_data;

  private->surface = surface;

  image->bpp    = BYTES_PER_PIXEL(format);
  image->depth  = BITS_PER_PIXEL(format);
  image->type   = type;
  image->visual = visual;
  image->width  = width;
  image->height = height;
  image->depth  = visual->depth;

  image->byte_order = GDK_LSB_FIRST;

  surface->Lock( surface, DSLF_WRITE, &image->mem, &pitch );
   
  image->bpl    = pitch;
   
  return image;
}

GdkImage*
_gdk_directfb_get_image (GdkDrawable *drawable,
                         gint         x,
                         gint         y,
                         gint         width,
                         gint         height)
{
  GdkImage                *image;
  GdkImageDirectFB        *private;
  GdkDrawableImplDirectFB *impl;
  
  DFBRectangle rect = { x, y, width, height };

  impl = GDK_DRAWABLE_IMPL_DIRECTFB (drawable);

  if (impl->wrapper == gdk_parent_root)
    {
      DFBResult ret;

      ret = DisplayLayer->SetCooperativeLevel (DisplayLayer, DLSCL_EXCLUSIVE);
      if (ret)
        {
          DirectFBError ("_gdk_directfb_get_image - SetCooperativeLevel", ret);
          return NULL;
        }

      ret = DisplayLayer->GetSurface (DisplayLayer, &impl->surface);
      if (ret)
        {
          DisplayLayer->SetCooperativeLevel (DisplayLayer, DLSCL_SHARED);
          DirectFBError ("_gdk_directfb_get_image - GetSurface", ret);
          return NULL;
        }
    }

  if (! impl->surface)
    return NULL;

  image = gdk_image_new (GDK_IMAGE_NORMAL,
                         gdk_visual_get_system (),
                         width, height);
  
  private = image->windowing_data;

  private->surface->Unlock( private->surface );
  private->surface->Blit( private->surface, impl->surface, &rect, 0, 0 );

  {
    int pitch;

    private->surface->Lock( private->surface, DSLF_WRITE,
                            &image->mem, &pitch );

    image->bpl = pitch;
  }

  if (impl->wrapper == gdk_parent_root)
    {
      impl->surface->Release (impl->surface);
      free (impl->surface);
      impl->surface = NULL;

      DisplayLayer->SetCooperativeLevel (DisplayLayer, DLSCL_SHARED);
    }

  return image;
}

guint32
gdk_image_get_pixel (GdkImage *image,
                     gint      x,
                     gint      y)
{
  guint32 pixel = 0;

  g_return_val_if_fail (GDK_IS_IMAGE (image), 0);

  if (!(x >= 0 && x < image->width && y >= 0 && y < image->height))
    return 0;

  if (image->depth == 1)
    pixel = (((guchar *) image->mem)[y * image->bpl + (x >> 3)] & (1 << (7 - (x & 0x7)))) != 0;
  else
    {
      guchar *pixelp = (guchar *) image->mem + y * image->bpl + x * image->bpp;
      
      switch (image->bpp)
        {
        case 1:
          pixel = *pixelp;
          break;

        case 2:
          pixel = pixelp[0] | (pixelp[1] << 8);
          break;

        case 3:
          pixel = pixelp[0] | (pixelp[1] << 8) | (pixelp[2] << 16);
          break;

        case 4:
          pixel = pixelp[0] | (pixelp[1] << 8) | (pixelp[2] << 16);
          break;
        }
    }

  return pixel;
}

void
gdk_image_put_pixel (GdkImage *image,
                     gint       x,
                     gint       y,
                     guint32    pixel)
{
  g_return_if_fail (image != NULL);

  g_message("%s", __FUNCTION__);

  if  (!(x >= 0 && x < image->width && y >= 0 && y < image->height))
    return;

  if (image->depth == 1)
    if (pixel & 1)
      ((guchar *) image->mem)[y * image->bpl + (x >> 3)] |= (1 << (7 - (x & 0x7)));
    else
      ((guchar *) image->mem)[y * image->bpl + (x >> 3)] &= ~(1 << (7 - (x & 0x7)));
  else
    {
      guchar *pixelp = (guchar *) image->mem + y * image->bpl + x * image->bpp;

      /* Windows is always LSB, no need to check image->byte_order. */
      switch (image->bpp)
        {
        case 4:
          pixelp[3] = 0;
        case 3:
          pixelp[2] = ((pixel >> 16) & 0xFF);
        case 2:
          pixelp[1] = ((pixel >> 8) & 0xFF);
        case 1:
          pixelp[0] = (pixel & 0xFF);
        }
    }
}

static void
gdk_directfb_image_destroy (GdkImage *image)
{
  GdkImageDirectFB *private;

  g_return_if_fail (GDK_IS_IMAGE (image));

  private = image->windowing_data;

  GDK_NOTE (MISC, g_print ("gdk_directfb_image_destroy: %#x\n",
                           (guint) private->surface));

  switch (image->type)
    {
    case GDK_IMAGE_SHARED:
      //TODODO
      break;

    default:
      g_assert_not_reached ();
    }

  private->surface->Unlock( private->surface );
  private->surface->Release( private->surface );
  free (private->surface);

  g_free (private);
  image->windowing_data = NULL;
}

