/* GDK - The GIMP Drawing Kit
 * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
 * Copyright (C) 1998-1999 Tor Lillqvist
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * Modified by the GTK+ Team and others 1997-2000.  See the AUTHORS
 * file for a list of people on the GTK+ Team.  See the ChangeLog
 * files for a list of changes.  These files are distributed with
 * GTK+ at ftp://ftp.gtk.org/pub/gtk/.
 */

#include "config.h"

#include <stdlib.h>

#include "gdkinternals.h"
#include "gdkprivate-directfb.h"
#include "gdkregion-generic.h"


static GdkRegion*  gdk_window_impl_directfb_get_visible_region (GdkDrawable *drawable);
static void gdk_window_impl_directfb_init       (GdkWindowImplDirectFB      *window);
static void gdk_window_impl_directfb_class_init (GdkWindowImplDirectFBClass *klass);
static void gdk_window_impl_directfb_finalize   (GObject                    *object);

void
gdk_window_invalidate_rect_clear (GdkWindow    *window,
                                  GdkRectangle *rect);

typedef struct
{
  GdkWindowChildChanged  changed;
  GdkWindowChildGetPos   get_pos;
  gpointer               user_data;
} GdkWindowChildHandlerData;


//static void gdk_directfb_window_expose( GdkWindow *window );

static GdkWindow *gdk_directfb_window_containing_pointer = NULL;
static GdkWindow *gdk_directfb_focused_window            = NULL;
static gpointer   parent_class                           = NULL;


GType
gdk_window_impl_directfb_get_type (void)
{
  static GType object_type = 0;

  if (!object_type)
    {
      static const GTypeInfo object_info =
        {
          sizeof (GdkWindowImplDirectFBClass),
          (GBaseInitFunc) NULL,
          (GBaseFinalizeFunc) NULL,
          (GClassInitFunc) gdk_window_impl_directfb_class_init,
          NULL,           /* class_finalize */
          NULL,           /* class_data */
          sizeof (GdkWindowImplDirectFB),
          0,              /* n_preallocs */
          (GInstanceInitFunc) gdk_window_impl_directfb_init,
        };
      
      object_type = g_type_register_static (GDK_TYPE_DRAWABLE_IMPL_DIRECTFB,
                                            "GdkWindowImplDirectFB",
                                            &object_info, 0);
    }
  
  return object_type;
}

GType
_gdk_window_impl_get_type (void)
{
  return gdk_window_impl_directfb_get_type ();
}

static void
gdk_window_impl_directfb_init (GdkWindowImplDirectFB *impl)
{
  impl->drawable.width = 1;
  impl->drawable.height = 1;

  impl->drawable.colormap = gdk_colormap_get_system ();
  impl->event_mask = 0;
  impl->cursor = gdk_cursor_new (GDK_LEFT_PTR);

  impl->opacity = 255;
}

static void
gdk_window_impl_directfb_class_init (GdkWindowImplDirectFBClass *klass)
{
  GObjectClass     *object_class   = G_OBJECT_CLASS (klass);
  GdkDrawableClass *drawable_class = GDK_DRAWABLE_CLASS (klass);

  parent_class = g_type_class_peek_parent (klass);

  object_class->finalize = gdk_window_impl_directfb_finalize;

  /* Visible and clip regions are the same */
  drawable_class->get_clip_region =
    gdk_window_impl_directfb_get_visible_region;

  drawable_class->get_visible_region =
    gdk_window_impl_directfb_get_visible_region;
}

static void
g_free_2nd (gpointer a,
            gpointer b,
            gpointer data)
{
  g_free (b);
}

static void
gdk_window_impl_directfb_finalize (GObject *object)
{
  GdkWindowImplDirectFB *impl = GDK_WINDOW_IMPL_DIRECTFB (object);

  if (GDK_WINDOW_IS_MAPPED (impl->drawable.wrapper))
    gdk_window_hide (impl->drawable.wrapper);

  if (impl->cursor)
    gdk_cursor_unref (impl->cursor);

  if (impl->properties)
    {
      g_hash_table_foreach (impl->properties, g_free_2nd, NULL);
      g_hash_table_destroy (impl->properties);
    }

  if (impl->window)
    {
      free (impl->window);
    }

  G_OBJECT_CLASS (parent_class)->finalize (object);
}

static GdkRegion*
gdk_window_impl_directfb_get_visible_region (GdkDrawable *drawable)
{
  GdkDrawableImplDirectFB *priv = GDK_DRAWABLE_IMPL_DIRECTFB (drawable);
  GdkRectangle             result_rect = { 0, 0, 0, 0 };

  if (priv->surface)
    {
      DFBRectangle rect;

      priv->surface->GetVisibleRectangle( priv->surface, &rect );

      result_rect.x      = rect.x;
      result_rect.y      = rect.y;
      result_rect.width  = rect.w;
      result_rect.height = rect.h;
    }

  return gdk_region_rectangle (&result_rect);
}

void
_gdk_windowing_window_init (void)
{
  GdkWindowObject       *private;
  GdkWindowImplDirectFB *impl;
  gint width;
  gint height;

  g_assert (gdk_parent_root == NULL);

  DisplayLayer->GetSize( DisplayLayer, &width, &height );

  gdk_parent_root = g_object_new (GDK_TYPE_WINDOW, NULL);
  private = GDK_WINDOW_OBJECT (gdk_parent_root);
  impl = GDK_WINDOW_IMPL_DIRECTFB (private->impl);

  private->window_type = GDK_WINDOW_ROOT;
  private->depth       = gdk_visual_get_system ()->depth;
  private->state       = 0;
  private->children    = NULL;

  impl->window           = NULL;
  impl->drawable.abs_x   = 0;
  impl->drawable.abs_y   = 0;
  impl->drawable.width   = width;
  impl->drawable.height  = height;
  impl->drawable.wrapper = GDK_DRAWABLE (private);
}

GdkWindow*
gdk_window_new (GdkWindow     *parent,
                GdkWindowAttr *attributes,
                gint           attributes_mask)
{
  GdkWindow             *window;
  GdkWindowObject       *private;
  GdkWindowObject       *parent_private;
  GdkVisual             *visual;
  GdkWindowImplDirectFB *impl;
  GdkWindowImplDirectFB *parent_impl;

  DFBResult              ret;
   
  gint x, y;

  g_return_val_if_fail (attributes != NULL, NULL);

  if (!parent || attributes->window_type != GDK_WINDOW_CHILD)
    parent = gdk_parent_root;

  window = GDK_WINDOW (g_object_new (GDK_TYPE_WINDOW, NULL));
  private = GDK_WINDOW_OBJECT (window);

  parent_private = GDK_WINDOW_OBJECT (parent);
  parent_impl = GDK_WINDOW_IMPL_DIRECTFB (parent_private->impl);
  private->parent = parent_private;

  if (attributes_mask & GDK_WA_X)
    x = attributes->x;
  else
    x = 0;

  if (attributes_mask & GDK_WA_Y)
    y = attributes->y;
  else
    y = 0;


  gdk_window_set_events( window, attributes->event_mask );

  if (attributes_mask & GDK_WA_VISUAL)
    visual = attributes->visual; //FIXME
  else
    visual = gdk_visual_get_system (); //FIXME


  impl = GDK_WINDOW_IMPL_DIRECTFB (private->impl);
  impl->drawable.wrapper = GDK_DRAWABLE (window);
  private->x = x;
  private->y = y;

  gdk_directfb_calc_abs (window);

  impl->drawable.width = (attributes->width > 1) ? (attributes->width) : (1);
  impl->drawable.height = (attributes->height > 1) ? (attributes->height) : (1);
  private->window_type = impl->drawable.window_type = attributes->window_type;

  if (attributes->wclass == GDK_INPUT_OUTPUT)
    {
      DFBWindowDescription desc;

      desc.flags = 0;
      private->input_only = FALSE;

      if ((attributes_mask & GDK_WA_COLORMAP) && attributes->colormap)
        impl->drawable.colormap = attributes->colormap;
      else
        impl->drawable.colormap = gdk_colormap_get_system();

      if (attributes_mask & GDK_WA_ALPHACHANNEL)
        {
          desc.flags |= DWDESC_CAPS;
          desc.caps = DWCAPS_ALPHACHANNEL;
        }

      switch (impl->drawable.window_type)
        {
        case GDK_WINDOW_TOPLEVEL:
        case GDK_WINDOW_DIALOG:
        case GDK_WINDOW_TEMP:
          desc.flags |= (DWDESC_WIDTH | DWDESC_HEIGHT |
                         DWDESC_POSX | DWDESC_POSY);
          desc.posx   = x;
          desc.posy   = y;
          desc.width  = impl->drawable.width;
          desc.height = impl->drawable.height;

          ret = DisplayLayer->CreateWindow (DisplayLayer,
                                            &desc, &impl->window);
          if (ret != DFB_OK)
            {
              DirectFBError( "gdk_window_new: Layer->CreateWindow",
                             ret );
              _gdk_window_destroy (window, FALSE);
              return NULL;
            }

          ret = impl->window->GetSurface( impl->window,
                                          &impl->drawable.surface );
          if (ret != DFB_OK)
            {
              DirectFBError ("gdk_window_new: Window->GetSurface",
                             ret);
              _gdk_window_destroy (window, FALSE);
              return NULL;
            }

          break;
        case GDK_WINDOW_CHILD:
          if (parent_impl->drawable.surface) {
            DFBRectangle rect;
            
            rect.x = x;
            rect.y = y;
            rect.w = impl->drawable.width;
            rect.h = impl->drawable.height;
            
            parent_impl->drawable.surface->GetSubSurface(
                                                         parent_impl->drawable.surface,
                                                         &rect, &impl->drawable.surface);
          }
          
          break;
        default:
          g_error ("unsupported window class");
          _gdk_window_destroy (window, FALSE);
          return NULL;
        }

      if (impl->drawable.surface)
        impl->drawable.surface->GetPixelFormat( impl->drawable.surface,
                                                &impl->drawable.format );
      else
        impl->drawable.format = DSPF_RGB16; //FIXME!!!
      
      private->depth = BITS_PER_PIXEL(impl->drawable.format);
    }
  else
    {
      private->input_only = TRUE;
      impl->drawable.colormap = NULL;
    }
  
  if (impl->drawable.colormap)
    gdk_colormap_ref (impl->drawable.colormap);

  gdk_window_set_cursor (window, ((attributes_mask & GDK_WA_CURSOR) ?
                                  (attributes->cursor) : NULL));

  if (parent_private)
    parent_private->children = g_list_prepend (parent_private->children, window);

  gdk_drawable_ref (window);

  if (impl->window)
    gdk_directfb_start_event_thread (window);

  return window;
}

void
_gdk_windowing_window_destroy (GdkWindow *window,
                               gboolean   recursing,
                               gboolean   foreign_destroy)
{
  GdkWindowObject       *private;
  GdkWindowImplDirectFB *impl;

  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  private = GDK_WINDOW_OBJECT (window);
  impl = GDK_WINDOW_IMPL_DIRECTFB (private->impl);

  _gdk_selection_window_destroyed (window);

  if (impl->window)
    impl->window->Release (impl->window);
}

/* This function is called when the window is eally gone.
 */
void
gdk_window_destroy_notify (GdkWindow *window)
{
  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  if (!GDK_WINDOW_DESTROYED (window))
    {
      if (GDK_WINDOW_TYPE(window) != GDK_WINDOW_FOREIGN)
	g_warning ("GdkWindow %p unexpectedly destroyed", window);

      _gdk_window_destroy (window, TRUE);
    }

  g_message ("%s: %p", __FUNCTION__, window);
  
  gdk_drawable_unref (window);
}

/* Focus follows pointer */
GdkWindow *
gdk_directfb_window_find_toplevel (GdkWindow *window)
{
  GdkWindowObject *priv = GDK_WINDOW_OBJECT (window);

  while (priv != GDK_WINDOW_OBJECT (gdk_parent_root))
    {
      if ((priv->parent == GDK_WINDOW_OBJECT (gdk_parent_root)) &&
          GDK_WINDOW_IS_MAPPED (priv))
        return GDK_WINDOW (priv);

      priv = priv->parent;
    }

  return gdk_parent_root;
}

GdkWindow *
gdk_directfb_window_find_focus (void)
{
  if (_gdk_directfb_keyboard_grab_window)
    return _gdk_directfb_keyboard_grab_window;

  if (!gdk_directfb_focused_window)
    gdk_directfb_focused_window = gdk_window_ref (gdk_parent_root);

  return gdk_directfb_focused_window;
}

void
gdk_directfb_change_focus (GdkWindow *new_focus_window)
{
  GdkEventFocus *event;
  GdkWindow     *old_win;
  GdkWindow     *new_win;
  GdkWindow     *event_win;

  g_return_if_fail (new_focus_window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (new_focus_window));

  /* No focus changes while the pointer is grabbed */
  if (_gdk_directfb_pointer_grab_window)
    return;

  old_win = gdk_directfb_focused_window;
  new_win = gdk_directfb_window_find_toplevel (new_focus_window);

  if (old_win != new_win)
    {
      if (old_win)
        {
          event_win =
            gdk_directfb_keyboard_event_window (old_win,
                                                GDK_FOCUS_CHANGE);
          if (event_win)
            {
              event = (GdkEventFocus *)gdk_event_make (event_win,
                                                       GDK_FOCUS_CHANGE,
                                                       TRUE);
              event->in = FALSE;
            }
        }
      
      event_win = gdk_directfb_keyboard_event_window (new_win,
                                                      GDK_FOCUS_CHANGE);
      if (event_win)
        {
          event = (GdkEventFocus *) gdk_event_make (event_win,
                                                    GDK_FOCUS_CHANGE,
                                                    TRUE);
          event->in = TRUE;
        }

      if (gdk_directfb_focused_window)
        gdk_window_unref (gdk_directfb_focused_window);

      gdk_directfb_focused_window = gdk_window_ref (new_win);
    }
}

static void
gdk_directfb_window_raise (GdkWindow *window)
{
  GdkWindowObject *parent;

  parent = GDK_WINDOW_OBJECT (window)->parent;

  parent->children = g_list_remove (parent->children, window);
  parent->children = g_list_prepend (parent->children, window);
}

static void
gdk_directfb_window_lower (GdkWindow *window)
{
  GdkWindowObject *parent;

  parent = GDK_WINDOW_OBJECT (window)->parent;

  parent->children = g_list_remove (parent->children, window);
  parent->children = g_list_append (parent->children, window);
}

/* Cut & paste versions of the stuff in gdkwindow.c,
   with the addition of clearing the newly exposed region. */
void
gdk_window_invalidate_region_clear (GdkWindow *window,
                                    GdkRegion *region)
{
  GdkWindowObject *private;
  gint i;

  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  if (GDK_WINDOW_DESTROYED (window))
    return;
  
  private = GDK_WINDOW_OBJECT (window);

  if (private->input_only || !GDK_WINDOW_IS_MAPPED (private))
    return;

  if (private->bg_pixmap != GDK_NO_BG)
    for (i = 0; i < region->numRects; i++)
      gdk_window_clear_area (window,
                             region->rects[i].x1,
                             region->rects[i].y1,
                             region->rects[i].x2 - region->rects[i].x1,
                             region->rects[i].y2 - region->rects[i].y1);

  gdk_window_invalidate_region (window, region, FALSE);

  {
    GList        *tmp_list;
    GdkRectangle  child_rect;
    GdkRegion    *child_region;

    tmp_list = private->children;
    while (tmp_list)
      {
        GdkWindowObject *child = tmp_list->data;

        tmp_list = tmp_list->next;

        if (!child->input_only)
          {
            gint width, height;

            gdk_drawable_get_size (GDK_DRAWABLE (child),
                                   &width, &height);

            child_rect.x      = child->x;
            child_rect.y      = child->y;
            child_rect.width  = width;
            child_rect.height = height;

            child_region = gdk_region_rectangle (&child_rect);
            gdk_region_intersect (child_region, region);

            if (!gdk_region_empty (child_region))
              {
                gdk_region_offset (child_region,
                                   - child_rect.x,
                                   - child_rect.y);
                gdk_window_invalidate_region_clear (GDK_WINDOW (child),
                                                    child_region);
              }
            
            gdk_region_destroy (child_region);
          }
      }
  }
}

void
gdk_window_invalidate_rect_clear (GdkWindow    *window,
                                  GdkRectangle *rect)
{
  GdkWindowObject *private;

  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  g_return_if_fail (rect != NULL);

  private = GDK_WINDOW_OBJECT (window);

  if (private->input_only || !GDK_WINDOW_IS_MAPPED (private))
    return;

  if (GDK_WINDOW_OBJECT (window)->bg_pixmap != GDK_NO_BG)
    gdk_window_clear_area (window,
                           rect->x, rect->y, rect->width, rect->height);

  gdk_window_invalidate_rect (window, rect, FALSE);

  {
    GList        *tmp_list;
    GdkRectangle  child_rect;
    GdkRectangle  new_rect;

    tmp_list = private->children;
    while (tmp_list)
      {
        GdkWindowObject       *child;
        GdkWindowImplDirectFB *impl;
        
        child = GDK_WINDOW_OBJECT (tmp_list->data);
        impl = GDK_WINDOW_IMPL_DIRECTFB (child->impl);

        tmp_list = tmp_list->next;

        if (!child->input_only)
          {
            gint width, height;

            width  = GDK_DRAWABLE_IMPL_DIRECTFB (impl)->width;
            height = GDK_DRAWABLE_IMPL_DIRECTFB (impl)->height;

            child_rect.x = child->x;
            child_rect.y = child->y;
            child_rect.width  = width;
            child_rect.height = height;

            if (gdk_rectangle_intersect (rect, &child_rect, &new_rect))
              {
                new_rect.x -= child_rect.x;
                new_rect.y -= child_rect.y;

                gdk_window_invalidate_rect_clear (GDK_WINDOW (child),
                                                  &new_rect);
              }
          }
      }
  }
}

static gboolean
all_parents_shown (GdkWindowObject *private)
{
  while (GDK_WINDOW_IS_MAPPED (private))
    {
      if (private->parent)
        private = GDK_WINDOW_OBJECT (private)->parent;
      else
        return TRUE;
    }

  return FALSE;
}

static void
send_map_events (GdkWindowObject *private,
                 gboolean         is_map)
{
  GdkWindow *parent = GDK_WINDOW (private->parent);
  GList     *list;
  GdkWindow *event_win;

  g_assert (is_map);

  if (!GDK_WINDOW_IS_MAPPED (private))
    return;

  if (is_map)
    {
      event_win =
        gdk_directfb_other_event_window ((GdkWindow *)private, GDK_MAP);
      if (event_win)
        gdk_event_make (event_win, GDK_MAP, TRUE);
    }

  if (private->input_only)
    return;

  if (!parent)
    parent = GDK_WINDOW (private);

  if (is_map)
    gdk_window_clear (GDK_WINDOW (private));

  for (list = private->children; list; list = list->next)
    send_map_events (list->data, is_map);
}

static GdkWindow *
gdk_directfb_find_common_ancestor (GdkWindow *win1,
                                   GdkWindow *win2)
{
  GdkWindowObject *tmp;
  GList *path1 = NULL;
  GList *path2 = NULL;
  GList *list1;
  GList *list2;

  tmp = GDK_WINDOW_OBJECT (win1);
  while (tmp)
    {
      path1 = g_list_prepend (path1, tmp);
      tmp = tmp->parent;
    }

  tmp = GDK_WINDOW_OBJECT (win2);
  while (tmp)
    {
      path2 = g_list_prepend (path2, tmp);
      tmp = tmp->parent;
    }

  list1 = path1;
  list2 = path2;
  tmp = NULL;
  while (list1 && list2 && (list1->data == list2->data))
    {
      tmp = GDK_WINDOW_OBJECT (list1->data);
      list1 = g_list_next (list1);
      list2 = g_list_next (list2);
    }

  g_list_free (path1);
  g_list_free (path2);

  return GDK_WINDOW (tmp);
}

void
gdk_directfb_window_send_crossing_events (GdkWindow       *src,
                                          GdkWindow       *dest,
                                          GdkCrossingMode  mode)
{
  GdkWindow       *c;
  GdkWindow       *win, *last, *next;
  GdkEvent        *event;
  gint             x, y, x_int, y_int;
  GdkModifierType  my_mask;
  GList           *path, *list;
  gboolean         non_linear;
  GdkWindow       *a;
  GdkWindow       *b;
  GdkWindow       *event_win;

  if ((mode == GDK_CROSSING_NORMAL) &&
      (dest == gdk_directfb_window_containing_pointer))
    return;

  if (gdk_directfb_window_containing_pointer == NULL)
    gdk_directfb_window_containing_pointer = gdk_window_ref (gdk_parent_root);

  if (src)
    a = src;
  else
    a = gdk_directfb_window_containing_pointer;

  b = dest;

  if (a == b)
    return;
  else
    {
      static GdkCursorDirectFB *last_cursor = NULL;

      GdkWindowImplDirectFB *impl = GDK_WINDOW_IMPL_DIRECTFB (
                                                   GDK_WINDOW_OBJECT (b)->impl);
      GdkCursorDirectFB *cursor = (GdkCursorDirectFB*) impl->cursor;

      if (cursor != last_cursor)
        {
          DisplayLayer->SetCursorShape (DisplayLayer, cursor->shape,
                                        cursor->hot_x, cursor->hot_y);
          last_cursor = cursor;
        }
    }

  /* gdk_directfb_window_containing_pointer might have been destroyed.
   * The refcount we hold on it should keep it, but it's parents
   * might have died.
   */
  if (GDK_WINDOW_DESTROYED (a))
    a = gdk_parent_root;

  gdk_directfb_mouse_get_info (&x, &y, &my_mask);

  c = gdk_directfb_find_common_ancestor (a, b);

  non_linear = (c != a) && (c != b);

  event_win = gdk_directfb_pointer_event_window (a, GDK_LEAVE_NOTIFY);
  if (event_win)
    {
      event = gdk_event_make (event_win, GDK_LEAVE_NOTIFY, TRUE);
      event->crossing.subwindow = NULL;

      gdk_window_get_origin (a, &x_int, &y_int);

      event->crossing.x      = x - x_int;
      event->crossing.y      = y - y_int;
      event->crossing.x_root = x;
      event->crossing.y_root = y;
      event->crossing.mode   = mode;

      if (non_linear)
        event->crossing.detail = GDK_NOTIFY_NONLINEAR;
      else if (c==a)
        event->crossing.detail = GDK_NOTIFY_INFERIOR;
      else
        event->crossing.detail = GDK_NOTIFY_ANCESTOR;

      event->crossing.focus = FALSE;
      event->crossing.state = my_mask;
    }

   /* Traverse up from a to (excluding) c */
  if (c != a)
    {
      last = a;
      win = GDK_WINDOW (GDK_WINDOW_OBJECT (a)->parent);
      while (win != c)
        {
          event_win =
            gdk_directfb_pointer_event_window (win, GDK_LEAVE_NOTIFY);

          if (event_win)
            {
              event = gdk_event_make (event_win,
                                      GDK_LEAVE_NOTIFY, TRUE);
              event->crossing.subwindow = gdk_window_ref (last);
              
              gdk_window_get_origin (win, &x_int, &y_int);

              event->crossing.x      = x - x_int;
              event->crossing.y      = y - y_int;
              event->crossing.x_root = x;
              event->crossing.y_root = y;
              event->crossing.mode   = mode;

              if (non_linear)
                event->crossing.detail = GDK_NOTIFY_NONLINEAR_VIRTUAL;
              else
                event->crossing.detail = GDK_NOTIFY_VIRTUAL;

              event->crossing.focus = FALSE;
              event->crossing.state = my_mask;
            }
          last = win;
          win = GDK_WINDOW (GDK_WINDOW_OBJECT (win)->parent);
        }
    }

  /* Traverse down from c to b */
  if (c != b)
    {
      path = NULL;
      win = GDK_WINDOW( GDK_WINDOW_OBJECT (b)->parent);
      while (win != c)
        {
          path = g_list_prepend (path, win);
          win = GDK_WINDOW( GDK_WINDOW_OBJECT (win)->parent);
        }

      list = path;
      while (list)
        {
          win = GDK_WINDOW (list->data);
          list = g_list_next (list);
          if (list)
            next = GDK_WINDOW (list->data);
          else
            next = b;

          event_win =
            gdk_directfb_pointer_event_window (win, GDK_ENTER_NOTIFY);

          if (event_win)
            {
              event = gdk_event_make (event_win,
                                      GDK_ENTER_NOTIFY, TRUE);
              event->crossing.subwindow = gdk_window_ref (next);
              
              gdk_window_get_origin (win, &x_int, &y_int);

              event->crossing.x      = x - x_int;
              event->crossing.y      = y - y_int;
              event->crossing.x_root = x;
              event->crossing.y_root = y;
              event->crossing.mode   = mode;

              if (non_linear)
                event->crossing.detail = GDK_NOTIFY_NONLINEAR_VIRTUAL;
              else
                event->crossing.detail = GDK_NOTIFY_VIRTUAL;

              event->crossing.focus = FALSE;
              event->crossing.state = my_mask;
            }
        }
      g_list_free (path);
    }

  event_win = gdk_directfb_pointer_event_window (b, GDK_ENTER_NOTIFY);
  if (event_win)
    {
      event = gdk_event_make (event_win, GDK_ENTER_NOTIFY, TRUE);
      event->crossing.subwindow = NULL;

      gdk_window_get_origin (b, &x_int, &y_int);

      event->crossing.x      = x - x_int;
      event->crossing.y      = y - y_int;
      event->crossing.x_root = x;
      event->crossing.y_root = y;
      event->crossing.mode   = mode;

      if (non_linear)
        event->crossing.detail = GDK_NOTIFY_NONLINEAR;
      else if (c==a)
        event->crossing.detail = GDK_NOTIFY_ANCESTOR;
      else
        event->crossing.detail = GDK_NOTIFY_INFERIOR;

      event->crossing.focus = FALSE;
      event->crossing.state = my_mask;
    }

  if (mode != GDK_CROSSING_GRAB)
    {
      gdk_directfb_change_focus (b);
      if (b != gdk_directfb_window_containing_pointer)
        {
          gdk_window_unref (gdk_directfb_window_containing_pointer);
          gdk_directfb_window_containing_pointer = gdk_window_ref (b);
        }
    }
}

void
gdk_window_show (GdkWindow *window)
{
  GdkWindowObject       *private;
  GdkWindowImplDirectFB *impl;
  GdkWindow             *mousewin;

  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  private = GDK_WINDOW_OBJECT (window);
  impl = GDK_WINDOW_IMPL_DIRECTFB (private->impl);

  if (!private->destroyed && !GDK_WINDOW_IS_MAPPED (private))
    {
      private->state &= ~GDK_WINDOW_STATE_WITHDRAWN;
      gdk_window_raise (window);

      if (all_parents_shown (GDK_WINDOW_OBJECT (private)->parent))
        {
          send_map_events (private, TRUE);

          mousewin = gdk_window_at_pointer (NULL, NULL);
          gdk_directfb_window_send_crossing_events (NULL, mousewin,
                                                    GDK_CROSSING_NORMAL);

          if (private->input_only)
            return;
          else
            {
              GdkRegion *reg;

              reg = gdk_window_impl_directfb_get_visible_region (GDK_DRAWABLE(impl));
              gdk_window_invalidate_region (window, reg, TRUE);
              gdk_region_destroy (reg);
            }
        }
    }
  
  if (impl->window)
    {
      if (gdk_directfb_apply_focus_opacity)
	impl->window->SetOpacity( impl->window,
				  (impl->opacity >> 1) +
				  (impl->opacity >> 2) );
      else
	impl->window->SetOpacity( impl->window, impl->opacity );
    }
}

void
gdk_window_hide (GdkWindow *window)
{
  GdkWindowObject       *private;
  GdkWindowImplDirectFB *impl;
  GdkWindow             *mousewin;
  GdkWindow             *event_win;

  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  private = GDK_WINDOW_OBJECT (window);
  impl = GDK_WINDOW_IMPL_DIRECTFB (private->impl);

  if (impl->window)
    impl->window->SetOpacity( impl->window, 0 );

  if (!private->destroyed && GDK_WINDOW_IS_MAPPED (private))
    {
      GdkEvent  *event;

      event_win = gdk_directfb_other_event_window (window, GDK_UNMAP);
      if (event_win)
        event = gdk_event_make (event_win, GDK_UNMAP, TRUE);

      private->state |= GDK_WINDOW_STATE_WITHDRAWN;

      mousewin = gdk_window_at_pointer (NULL, NULL);
      gdk_directfb_window_send_crossing_events (NULL,
                                                mousewin,
                                                GDK_CROSSING_NORMAL);
      
      if (window == _gdk_directfb_pointer_grab_window)
        gdk_pointer_ungrab (GDK_CURRENT_TIME);
      if (window == _gdk_directfb_keyboard_grab_window)
        gdk_keyboard_ungrab (GDK_CURRENT_TIME);

      /* Invalidate the region */
      if (!impl->window)
        {
          GdkRegion *reg;

          reg = gdk_window_impl_directfb_get_visible_region (GDK_DRAWABLE(impl));

          gdk_region_offset (reg, private->x, private->y);

          gdk_window_invalidate_region (GDK_WINDOW (private->parent),
                                        reg, TRUE);
          
          gdk_region_destroy (reg);
        }
    }
}

void
gdk_window_withdraw (GdkWindow *window)
{
  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));
  
  /* for now this should be enough */
  gdk_window_hide (window);
}

void
gdk_window_move (GdkWindow *window,
                 gint       x,
                 gint       y)
{
  GdkWindowObject *private;
  GdkWindowImplDirectFB *impl;
  
  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));
  
  private = GDK_WINDOW_OBJECT (window);
  impl = GDK_WINDOW_IMPL_DIRECTFB (private->impl);
  
  if (impl->window)
    {
      impl->window->MoveTo( impl->window, x, y );
    }
  else
    {
      int                    dx, dy;
      GdkWindowObject       *parent;
      GdkWindowImplDirectFB *parent_impl;
      IDirectFBSurface      *parent_surface;
      GdkRegion             *child_region_before = NULL;
      GdkRegion             *child_region_after = NULL;
      GdkRectangle           child_rectangle;

      DFBRectangle  rect;
      DFBRegion     update;

      parent = GDK_WINDOW_OBJECT (private->parent);
      parent_impl = GDK_WINDOW_IMPL_DIRECTFB (parent->impl);
      parent_surface = parent_impl->drawable.surface;

      dx = x - private->x;
      dy = y - private->y;

      rect.w = rect.h = 0;

      if (impl->drawable.surface)
	{
	  impl->drawable.surface->GetVisibleRectangle (impl->drawable.surface, &rect);
	}

      if (rect.w > 0  &&  rect.h > 0)
	{
	  rect.x += private->x;
	  rect.y += private->y;

	  child_rectangle.x = rect.x;
	  child_rectangle.y = rect.y;
	  child_rectangle.width = rect.w;
	  child_rectangle.height = rect.h;
	  child_region_before = gdk_region_rectangle (&child_rectangle);

	  parent_surface->Blit (parent_surface, parent_surface, &rect, rect.x + dx, rect.y + dy);


	  rect.x += dx;
	  rect.y += dy;

	  child_rectangle.x = rect.x;
	  child_rectangle.y = rect.y;
	  child_region_after = gdk_region_rectangle (&child_rectangle);

	  update.x1 = rect.x;
	  update.y1 = rect.y;
	  update.x2 = rect.x + rect.w - 1;
	  update.y2 = rect.y + rect.h - 1;
	  
	  gdk_directfb_add_update (&parent_impl->drawable, &update);
	}
      else
	{
	  child_rectangle.x = x;
	  child_rectangle.y = y;
	  child_rectangle.width = impl->drawable.width;
	  child_rectangle.height = impl->drawable.height;
	  child_region_after = gdk_region_rectangle (&child_rectangle);
	}

      gdk_directfb_move_resize_child (window, x, y, impl->drawable.width, impl->drawable.height);
      gdk_directfb_calc_abs (window);

      if (child_region_before && child_region_after)
	{
	  gdk_region_subtract (child_region_before, child_region_after);

	  gdk_window_invalidate_region (GDK_WINDOW (parent), child_region_before, TRUE);

	  gdk_region_destroy (child_region_before);
	  gdk_region_destroy (child_region_after);
	}
      else
	{
	  gdk_window_invalidate_region (GDK_WINDOW (parent), child_region_after, TRUE);

	  gdk_region_destroy (child_region_after);
	}
    }

}

void
gdk_window_resize (GdkWindow *window,
                   gint       width,
                   gint       height)
{
  GdkWindowObject *private;
  gint x, y;

  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  private = GDK_WINDOW_OBJECT (window);

  if (width < 1)
    width = 1;
  if (height < 1)
    height = 1;

  x = private->x;
  y = private->y;

  if (private->parent && (private->parent->window_type != GDK_WINDOW_CHILD))
    {
      GdkWindowChildHandlerData *data;

      data = g_object_get_data (G_OBJECT (private->parent),
                                "gdk-window-child-handler");

      if (data)
        (*data->get_pos) (window, &x, &y, data->user_data);
    }

  gdk_window_move_resize (window, x, y, width, height);
}

void
gdk_directfb_move_resize_child (GdkWindow *window,
                                gint       x,
                                gint       y,
                                gint       width,
                                gint       height)
{
  GdkWindowObject       *private;
  GdkWindowImplDirectFB *impl;
  GdkWindowImplDirectFB *parent_impl;
  GList                 *list;

  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  private = GDK_WINDOW_OBJECT (window);
  impl = GDK_WINDOW_IMPL_DIRECTFB (private->impl);

  private->x = x;
  private->y = y;

  impl->drawable.width  = width;
  impl->drawable.height = height;

  if (!private->input_only)
    {
      if (impl->drawable.surface)
        {
          impl->drawable.surface->Release( impl->drawable.surface );
          free (impl->drawable.surface);
          impl->drawable.surface = NULL;
        }

      parent_impl = GDK_WINDOW_IMPL_DIRECTFB (GDK_WINDOW_OBJECT (private->parent)->impl);
      
      if (parent_impl->drawable.surface)
        {
          DFBRectangle rect;
          
          rect.x = x;
          rect.y = y;
          rect.w = width;
          rect.h = height;
          
          parent_impl->drawable.surface->GetSubSurface (
                                                        parent_impl->drawable.surface,
                                                        &rect, &impl->drawable.surface );
        }
    }
  
  for (list = private->children; list; list = list->next)
    {
      GdkWindowObject         *win = GDK_WINDOW_OBJECT (list->data);
      GdkDrawableImplDirectFB *i   = GDK_DRAWABLE_IMPL_DIRECTFB (win->impl);

      gdk_directfb_move_resize_child (GDK_WINDOW (list->data),
                                      win->x, win->y, i->width, i->height);
    }
}

void
gdk_window_move_resize (GdkWindow *window,
                        gint       x,
                        gint       y,
                        gint       width,
                        gint       height)
{
  GdkWindowObject       *private;
  GdkWindowImplDirectFB *impl;

  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  private = GDK_WINDOW_OBJECT (window);
  impl = GDK_WINDOW_IMPL_DIRECTFB (private->impl);
  
  if (width < 1)
     width = 1;
  if (height < 1)
    height = 1;

  if (private->destroyed ||
      (private->x == x  &&  private->y == y  &&
       impl->drawable.width == width  &&  impl->drawable.height == height))
    return;

  if (private->parent && (private->parent->window_type != GDK_WINDOW_CHILD))
    {
      GdkWindowChildHandlerData *data;

      data = g_object_get_data (G_OBJECT (private->parent),
                                "gdk-window-child-handler");

      if (data && (*data->changed) (window, x, y, width, height,
                                    data->user_data))
        return;
    }


  if (impl->window)
    {
      impl->window->Resize( impl->window, width, height );
      impl->window->MoveTo( impl->window, x, y );
    }
  else if (impl->drawable.width == width  &&  impl->drawable.height == height)
    gdk_window_move (window, x, y);
  else
    {
      GdkRectangle  rect;

      rect.x = private->x;
      rect.y = private->y;
      rect.width  = impl->drawable.width;
      rect.height = impl->drawable.height;

      gdk_directfb_move_resize_child (window, x, y, width, height);
      gdk_directfb_calc_abs (window);

      if (GDK_WINDOW_IS_MAPPED (private))
        {
          GdkWindow *mousewin;

          if (rect.width < width)
            rect.width = width;

          if (rect.height < height)
            rect.height = height;

          if (x < rect.x)
            {
              rect.width += rect.x - x;
              rect.x = x;
            }
          else
            rect.width += x - rect.x;
          
          if (y < rect.y)
            {
              rect.height += rect.y - y;
              rect.y = y;
            }
          else
            rect.width += y - rect.y;

          gdk_window_invalidate_rect (GDK_WINDOW (private->parent),
                                      &rect, TRUE);

          /* The window the pointer is in might have changed */
          mousewin = gdk_window_at_pointer (NULL, NULL);
          gdk_directfb_window_send_crossing_events (NULL,
                                                    mousewin,
                                                    GDK_CROSSING_NORMAL);
        }
    }
}

void
gdk_window_reparent (GdkWindow *window,
                     GdkWindow *new_parent,
                     gint       x,
                     gint       y)
{
  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  if (GDK_WINDOW_DESTROYED (window))
    return;
  
  /* N/A */
  g_message("unimplemented %s", __FUNCTION__);
}

void
_gdk_windowing_window_clear_area (GdkWindow *window,
                                  gint       x,
                                  gint       y,
                                  gint       width,
                                  gint       height)
{
  GdkWindowObject         *private;
  GdkDrawableImplDirectFB *impl;
  GdkPixmap               *bg_pixmap;
  GdkWindow               *relative_to;

  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  if (GDK_WINDOW_DESTROYED (window))
    return;

  private = GDK_WINDOW_OBJECT (window);
  impl    = GDK_DRAWABLE_IMPL_DIRECTFB (private->impl);

  if (width == 0)
    width = impl->width - x;
  if (height == 0)
    height = impl->height - y;

  bg_pixmap = private->bg_pixmap;

  for (relative_to = window;
       bg_pixmap == GDK_PARENT_RELATIVE_BG && relative_to;
       relative_to = (GdkWindow *) GDK_WINDOW_OBJECT (relative_to)->parent)
    {
      bg_pixmap = GDK_WINDOW_OBJECT (relative_to)->bg_pixmap;
    }

  if (bg_pixmap && bg_pixmap != GDK_NO_BG)
    {
      g_message ("%s(): unimplemented GDK_PARENT_RELATIVE_BG", G_GNUC_FUNCTION);
    }
  else if (! bg_pixmap)
    {
      gdk_directfb_draw_rectangle (GDK_DRAWABLE(impl),
				   NULL, TRUE, x, y, width, height);
    }
}

void
_gdk_windowing_window_clear_area_e (GdkWindow *window,
                                    gint       x,
                                    gint       y,
                                    gint       width,
                                    gint       height)
{
  g_message ("partly implemented %s", __FUNCTION__);

  _gdk_windowing_window_clear_area (window, x, y, width, height);
}

void
gdk_window_raise (GdkWindow *window)
{
  GdkWindowImplDirectFB *impl;

  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  if (GDK_WINDOW_DESTROYED (window))
    return;

  impl = GDK_WINDOW_IMPL_DIRECTFB (GDK_WINDOW_OBJECT (window)->impl);

  if (impl->window)
    {
      DFBResult ret;

      ret = impl->window->RaiseToTop (impl->window);
      if (ret)
        DirectFBError ("gdkwindow-directfb.c: RaiseToTop", ret);
      else
        gdk_directfb_window_raise (window);
    }
  else
    {
      gdk_directfb_window_raise (window);
      gdk_window_invalidate_rect (window, NULL, TRUE);
    }
}

void
gdk_window_lower (GdkWindow *window)
{
  GdkWindowImplDirectFB *impl;

  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  if (GDK_WINDOW_DESTROYED (window))
    return;

  impl = GDK_WINDOW_IMPL_DIRECTFB (GDK_WINDOW_OBJECT (window)->impl);

  if (impl->window)
    {
      DFBResult ret;

      ret = impl->window->LowerToBottom (impl->window);
      if (ret)
        DirectFBError ("gdkwindow-directfb.c: LowerToBottom", ret);
      else
        gdk_directfb_window_lower (window);
    }
  else
    {
      gdk_directfb_window_lower (window);
      gdk_window_invalidate_rect (window, NULL, TRUE);
    }
}

void
gdk_window_set_hints (GdkWindow *window,
                      gint       x,
                      gint       y,
                      gint       min_width,
                      gint       min_height,
                      gint       max_width,
                      gint       max_height,
                      gint       flags)
{
  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  if (GDK_WINDOW_DESTROYED (window))
    return;
  
  /* N/A */
  g_message("unimplemented %s", __FUNCTION__);
}

void
gdk_window_set_geometry_hints (GdkWindow      *window,
                               GdkGeometry    *geometry,
                               GdkWindowHints  geom_mask)
{
  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  if (GDK_WINDOW_DESTROYED (window))
    return;
  
  /* N/A */
  g_message("unimplemented %s", __FUNCTION__);
}

void
gdk_window_set_title (GdkWindow   *window,
                      const gchar *title)
{
  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  if (GDK_WINDOW_DESTROYED (window))
    return;
  
  /* N/A */
  g_message("unimplemented %s", __FUNCTION__);
}

void
gdk_window_set_role (GdkWindow   *window,
                     const gchar *role)
{
  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  if (GDK_WINDOW_DESTROYED (window))
    return;
  
  /* N/A */
  g_message("unimplemented %s", __FUNCTION__);
}

void
gdk_window_set_transient_for (GdkWindow *window,
                              GdkWindow *parent)
{
  GdkWindowObject *private;
  GdkWindowObject *root;
  gint i;

  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  g_return_if_fail (parent != NULL);
  g_return_if_fail (GDK_IS_WINDOW (parent));

  private = GDK_WINDOW_OBJECT (window);
  root    = GDK_WINDOW_OBJECT (gdk_parent_root);

  g_return_if_fail (GDK_WINDOW (private->parent) == gdk_parent_root);
  g_return_if_fail (GDK_WINDOW (GDK_WINDOW_OBJECT (parent)->parent) == gdk_parent_root);

  root->children = g_list_remove (root->children, window);
   
  i = g_list_index (root->children, parent);
  if (i < 0)
    root->children = g_list_prepend (root->children, window);
  else
    root->children = g_list_insert (root->children, window, i);
}

void
gdk_window_set_background (GdkWindow *window,
                           GdkColor  *color)
{
  GdkWindowObject *private;

  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  g_return_if_fail (color != NULL);

  private = GDK_WINDOW_OBJECT (window);
  private->bg_color = *color;

  if (private->bg_pixmap &&
      private->bg_pixmap != GDK_PARENT_RELATIVE_BG &&
      private->bg_pixmap != GDK_NO_BG)
    {
      gdk_pixmap_unref (private->bg_pixmap);
      private->bg_pixmap = NULL;
    }
}

void
gdk_window_set_back_pixmap (GdkWindow *window,
                            GdkPixmap *pixmap,
                            gint       parent_relative)
{
  GdkWindowObject *private;
  GdkPixmap       *old_pixmap;

  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));
  g_return_if_fail (pixmap == NULL || !parent_relative);

  private = GDK_WINDOW_OBJECT (window);
  old_pixmap = private->bg_pixmap;

  if (private->bg_pixmap &&
      private->bg_pixmap != GDK_PARENT_RELATIVE_BG &&
      private->bg_pixmap != GDK_NO_BG)
    gdk_pixmap_unref (private->bg_pixmap);

  if (parent_relative)
    {
      private->bg_pixmap = GDK_PARENT_RELATIVE_BG;
    }
  else
    {
      if (pixmap)
	{
	  gdk_pixmap_ref (pixmap);
	  private->bg_pixmap = pixmap;
	}
      else
	{
	  private->bg_pixmap = GDK_NO_BG;
	}
    }
}

void
gdk_window_set_cursor (GdkWindow *window,
                       GdkCursor *cursor)
{
  GdkWindowImplDirectFB *impl;
  GdkCursor             *old_cursor;

  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  impl = GDK_WINDOW_IMPL_DIRECTFB (GDK_WINDOW_OBJECT (window)->impl);
  old_cursor = impl->cursor;

  impl->cursor = cursor ? gdk_cursor_ref (cursor) :
    gdk_cursor_new (GDK_LEFT_PTR);

  if (old_cursor)
    gdk_cursor_unref (old_cursor);

  if (gdk_window_at_pointer (NULL, NULL) == window)
    {
      GdkCursorDirectFB *private = (GdkCursorDirectFB*) impl->cursor;

      DisplayLayer->SetCursorShape (DisplayLayer, private->shape,
                                    private->hot_x, private->hot_y);
    }
}

void
gdk_window_get_geometry (GdkWindow *window,
                         gint      *x,
                         gint      *y,
                         gint      *width,
                         gint      *height,
                         gint      *depth)
{
  GdkWindowObject         *private;
  GdkDrawableImplDirectFB *impl;

  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  private = GDK_WINDOW_OBJECT (window);
  impl    = GDK_DRAWABLE_IMPL_DIRECTFB (private->impl);

  if (!GDK_WINDOW_DESTROYED (window))
    {
      if (x)
	*x = private->x;

      if (y)
	*y = private->y;

      if (width)
	*width = impl->width;

      if (height)
	*height = impl->height;

      if (depth)
	*depth = BITS_PER_PIXEL(impl->format);
    }
}

void
gdk_directfb_calc_abs (GdkWindow *window)
{
  GdkWindowObject         *private;
  GdkDrawableImplDirectFB *impl;
  GList                   *list;

  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  private = GDK_WINDOW_OBJECT (window);
  impl = GDK_DRAWABLE_IMPL_DIRECTFB (private->impl);

  impl->abs_x = private->x;
  impl->abs_y = private->y;

  if (private->parent)
    {
      GdkDrawableImplDirectFB *parent_impl = GDK_DRAWABLE_IMPL_DIRECTFB (
                                                    GDK_WINDOW_OBJECT
                                                      (private->parent)->impl);

      impl->abs_x += parent_impl->abs_x;
      impl->abs_y += parent_impl->abs_y;
    }


  for (list = private->children; list; list = list->next)
    {
      gdk_directfb_calc_abs (list->data);
    }
}

gboolean
gdk_window_get_origin (GdkWindow *window,
                       gint      *x,
                       gint      *y)
{
  g_return_val_if_fail (window != NULL, FALSE);
  g_return_val_if_fail (GDK_IS_WINDOW (window), FALSE);

  if (!GDK_WINDOW_DESTROYED (window))
    {
      GdkDrawableImplDirectFB *impl;

      impl = GDK_DRAWABLE_IMPL_DIRECTFB (GDK_WINDOW_OBJECT (window)->impl);

      if (x)
	*x = impl->abs_x;
      if (y)
	*y = impl->abs_y;

      return TRUE;
    }

  return FALSE;
}

gboolean
gdk_window_get_deskrelative_origin (GdkWindow *window,
                                    gint      *x,
                                    gint      *y)
{
  return gdk_window_get_origin (window, x, y);
}

void
gdk_window_get_root_origin (GdkWindow *window,
                            gint      *x,
                            gint      *y)
{
  GdkWindowObject *rover;

  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  rover = (GdkWindowObject*) window;
  if (x)
    *x = 0;
  if (y)
    *y = 0;

  if (GDK_WINDOW_DESTROYED (window))
    return;
      
  while (rover->parent && ((GdkWindowObject*) rover->parent)->parent)
    rover = (GdkWindowObject *) rover->parent;
  if (rover->destroyed)
    return;

  if (x)
    *x = rover->x;
  if (y)
    *y = rover->y;
}

GdkWindow *
gdk_window_get_pointer (GdkWindow       *window,
                        gint            *x,
                        gint            *y,
                        GdkModifierType *mask)
{
  GdkWindow               *retval = NULL;
  gint                     rx, ry, wx, wy;
  GdkDrawableImplDirectFB *impl;

  g_return_val_if_fail (window == NULL || GDK_IS_WINDOW (window), NULL);

  if (!window)
    window = gdk_parent_root;

  gdk_directfb_mouse_get_info (&rx, &ry, mask);

  wx = rx;
  wy = ry;
  retval = gdk_directfb_child_at (gdk_parent_root, &wx, &wy);

  impl = GDK_DRAWABLE_IMPL_DIRECTFB (GDK_WINDOW_OBJECT (window)->impl);

  if (x)
    *x = rx - impl->abs_x;
  if (y)
    *y = ry - impl->abs_y;

  return retval;
}

GdkWindow*
gdk_window_at_pointer (gint *win_x,
                       gint *win_y)
{
  GdkWindow *retval;
  gint       wx, wy;

  gdk_directfb_mouse_get_info (&wx, &wy, NULL);

  if (win_x)
    wx = *win_x;

  if (win_y)
    wy = *win_y;

  retval = gdk_directfb_child_at (gdk_parent_root, &wx, &wy);

  if (win_x)
    *win_x = wx;

  if (win_y)
    *win_y = wy;

  return retval;
}

GdkEventMask
gdk_window_get_events (GdkWindow *window)
{
  g_return_val_if_fail (window != NULL, 0);
  g_return_val_if_fail (GDK_IS_WINDOW (window), 0);

  if (GDK_WINDOW_DESTROYED (window))
    return 0;
  else
    return GDK_WINDOW_IMPL_DIRECTFB (GDK_WINDOW_OBJECT (window)->impl)->event_mask;
}

void
gdk_window_set_events (GdkWindow    *window,
                       GdkEventMask  event_mask)
{
  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  if (event_mask & GDK_BUTTON_MOTION_MASK)
    event_mask |= GDK_BUTTON1_MOTION_MASK | GDK_BUTTON2_MOTION_MASK |
      GDK_BUTTON3_MOTION_MASK;
  
  if (!GDK_WINDOW_DESTROYED (window))
    GDK_WINDOW_IMPL_DIRECTFB (GDK_WINDOW_OBJECT (window)->impl)->event_mask = event_mask;
}

void
gdk_window_shape_combine_mask (GdkWindow *window,
                               GdkBitmap *mask,
                               gint       x, 
                               gint       y)
{
  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  if (GDK_WINDOW_DESTROYED (window))
    return;
  
  /* N/A */
  g_message ("unimplemented %s", __FUNCTION__);
}

void
gdk_window_shape_combine_region (GdkWindow *window,
                                 GdkRegion *shape_region,
                                 gint       offset_x,
                                 gint       offset_y)
{
  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  if (GDK_WINDOW_DESTROYED (window))
    return;
  
  /* N/A */
  g_message("unimplemented %s", __FUNCTION__);
}

void
gdk_window_set_override_redirect (GdkWindow *window,
                                  gboolean   override_redirect)
{
  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));
  
  if (GDK_WINDOW_DESTROYED (window))
    return;
  
  /* N/A */
  g_message("unimplemented %s", __FUNCTION__);
}

void
gdk_window_set_icon (GdkWindow *window,
                     GdkWindow *icon_window,
                     GdkPixmap *pixmap,
                     GdkBitmap *mask)
{
  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  if (GDK_WINDOW_DESTROYED (window))
    return;
  
  /* N/A */
  g_message("unimplemented %s", __FUNCTION__);
}

void
gdk_window_set_icon_name (GdkWindow   *window,
                          const gchar *name)
{
  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  if (GDK_WINDOW_DESTROYED (window))
    return;
  
  /* N/A */
  g_message("unimplemented %s", __FUNCTION__);
}

void
gdk_window_iconify (GdkWindow *window)
{
  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  if (GDK_WINDOW_DESTROYED (window))
    return;
  
  gdk_window_hide (window);
}

void
gdk_window_deiconify (GdkWindow *window)
{
  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  if (GDK_WINDOW_DESTROYED (window))
    return;

  gdk_window_show (window);
}

void
gdk_window_stick (GdkWindow *window)
{
  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  if (GDK_WINDOW_DESTROYED (window))
    return;
  
  /* N/A */
  g_message("unimplemented %s", __FUNCTION__);
}

void
gdk_window_unstick (GdkWindow *window)
{
  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  if (GDK_WINDOW_DESTROYED (window))
    return;
  
  /* N/A */
  g_message("unimplemented %s", __FUNCTION__);
}

void
gdk_window_set_opacity (GdkWindow *window,
			gdouble    opacity)
{
  GdkWindowImplDirectFB *impl;

  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  if (GDK_WINDOW_DESTROYED (window))
    return;

  impl = GDK_WINDOW_IMPL_DIRECTFB (GDK_WINDOW_OBJECT (window)->impl);

  impl->opacity = (guint8) (CLAMP (opacity, 0.0, 1.0) * 255.999);

  if (impl->window && GDK_WINDOW_IS_MAPPED (window))
    {
      if (gdk_directfb_apply_focus_opacity &&
	  window == gdk_directfb_focused_window)
	impl->window->SetOpacity( impl->window,
				  (impl->opacity >> 1) +
				  (impl->opacity >> 2) );
      else
	impl->window->SetOpacity( impl->window, impl->opacity );
    }
}

void
gdk_window_focus (GdkWindow *window,
                  guint32    timestamp)
{
  GdkWindow *toplevel;

  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  if (GDK_WINDOW_DESTROYED (window))
    return;
  
  toplevel = gdk_directfb_window_find_toplevel (window);
  if (toplevel != gdk_parent_root)
    {
      GdkWindowImplDirectFB *impl;

      impl = GDK_WINDOW_IMPL_DIRECTFB (GDK_WINDOW_OBJECT (toplevel)->impl);

      impl->window->RequestFocus (impl->window);
    }
}

void
gdk_window_maximize (GdkWindow *window)
{
  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  if (GDK_WINDOW_DESTROYED (window))
    return;
  
  /* N/A */
  g_message("unimplemented %s", __FUNCTION__);
}

void
gdk_window_unmaximize (GdkWindow *window)
{
  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  if (GDK_WINDOW_DESTROYED (window))
    return;
  
  /* N/A */
  g_message("unimplemented %s", __FUNCTION__);
}

void
gdk_window_set_type_hint (GdkWindow        *window,
                          GdkWindowTypeHint hint)
{
  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  if (GDK_WINDOW_DESTROYED (window))
    return;
  
  /* N/A */
  g_message("unimplemented %s", __FUNCTION__);
}

void
gdk_window_set_modal_hint (GdkWindow *window,
                           gboolean   modal)
{
  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  if (GDK_WINDOW_DESTROYED (window))
    return;
  
  /* N/A */
  g_message("unimplemented %s", __FUNCTION__);
}

void
gdk_window_set_group (GdkWindow *window,
                      GdkWindow *leader)
{
  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));
  g_return_if_fail (leader != NULL);
  g_return_if_fail (GDK_IS_WINDOW (leader));

  if (GDK_WINDOW_DESTROYED (window))
    return;
  
  /* N/A */
  g_message("unimplemented %s", __FUNCTION__);
}

void
gdk_fb_window_set_child_handler (GdkWindow             *window,
                                 GdkWindowChildChanged  changed,
                                 GdkWindowChildGetPos   get_pos,
                                 gpointer               user_data)
{
  GdkWindowChildHandlerData *data;

  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  data = g_new (GdkWindowChildHandlerData, 1);
  data->changed   = changed;
  data->get_pos   = get_pos;
  data->user_data = user_data;

  g_object_set_data_full (G_OBJECT (window), "gdk-window-child-handler",
                          data, (GDestroyNotify) g_free);
}

void
gdk_window_set_decorations (GdkWindow       *window,
                            GdkWMDecoration  decorations)
{
  GdkWMDecoration *dec;

  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  dec = g_new (GdkWMDecoration, 1);
  *dec = decorations;

  g_object_set_data_full (G_OBJECT (window), "gdk-window-decorations",
                          dec, (GDestroyNotify) g_free);
}

gboolean
gdk_window_get_decorations (GdkWindow       *window,
			    GdkWMDecoration *decorations)
{
  GdkWMDecoration *dec;

  g_return_val_if_fail (window != NULL, FALSE);
  g_return_val_if_fail (GDK_IS_WINDOW (window), FALSE);

  dec = g_object_get_data (G_OBJECT (window), "gdk-window-decorations");
  if (dec)
    {
      *decorations = *dec;
      return TRUE;
    }
  return FALSE;
}

void
gdk_window_set_functions (GdkWindow     *window,
                          GdkWMFunction  functions)
{
  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  if (GDK_WINDOW_DESTROYED (window))
    return;
  
  /* N/A */
  g_message("unimplemented %s", __FUNCTION__);
}

void
gdk_window_set_child_shapes (GdkWindow *window)
{
  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  if (GDK_WINDOW_DESTROYED (window))
    return;
  
  /* N/A */
  g_message("unimplemented %s", __FUNCTION__);
}

void
gdk_window_merge_child_shapes (GdkWindow *window)
{
  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  if (GDK_WINDOW_DESTROYED (window))
    return;
  
  /* N/A */
  g_message("unimplemented %s", __FUNCTION__);
}

gboolean
gdk_window_set_static_gravities (GdkWindow *window,
                                 gboolean   use_static)
{
  g_return_val_if_fail (window != NULL, FALSE);
  g_return_val_if_fail (GDK_IS_WINDOW (window), FALSE);

  if (GDK_WINDOW_DESTROYED (window))
    return FALSE;
  
  /* N/A */
  g_message("unimplemented %s", __FUNCTION__);

  return FALSE;
}

void
gdk_window_begin_paint_region (GdkWindow *window,
                               GdkRegion *region)
{
  GdkDrawableImplDirectFB *impl;

  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  impl = GDK_DRAWABLE_IMPL_DIRECTFB (GDK_WINDOW_OBJECT (window)->impl);

  impl->buffered = TRUE;

  if (!region)
    return;

  if (impl->paint_region)
    {
      g_warning ("FIXME: nested call to gdk_window_begin_paint_region()\n");
      gdk_region_union (impl->paint_region, region);
    }
  else
    impl->paint_region = gdk_region_copy (region);
}

void
gdk_window_end_paint (GdkWindow *window)
{
  GdkDrawableImplDirectFB *impl;

  g_return_if_fail (window != NULL);
  g_return_if_fail (GDK_IS_WINDOW (window));

  impl = GDK_DRAWABLE_IMPL_DIRECTFB (GDK_WINDOW_OBJECT (window)->impl);

  impl->buffered = FALSE;
  
  if (impl->paint_region)
    gdk_region_destroy (impl->paint_region);
  impl->paint_region = NULL;

  gdk_directfb_add_update (impl, NULL);
}


void
gdk_window_begin_resize_drag (GdkWindow     *window,
                              GdkWindowEdge  edge,
                              gint           button,
                              gint           root_x,
                              gint           root_y,
                              guint32        timestamp)
{
  g_return_if_fail (GDK_IS_WINDOW (window));
  
  if (GDK_WINDOW_DESTROYED (window))
    return;

  g_message("unimplemented %s", __FUNCTION__);
}

void
gdk_window_begin_move_drag (GdkWindow *window,
                            gint       button,
                            gint       root_x,
                            gint       root_y,
                            guint32    timestamp)
{
  g_return_if_fail (GDK_IS_WINDOW (window));
  
  if (GDK_WINDOW_DESTROYED (window))
    return;

  g_message("unimplemented %s", __FUNCTION__);
}

/**
 * gdk_window_get_frame_extents:
 * @window: a #GdkWindow
 * @rect: rectangle to fill with bounding box of the window frame
 *
 * Obtains the bounding box of the window, including window manager
 * titlebar/borders if any. The frame position is given in root window
 * coordinates. To get the position of the window itself (rather than
 * the frame) in root window coordinates, use gdk_window_get_origin().
 * 
 **/
void
gdk_window_get_frame_extents (GdkWindow    *window,
                              GdkRectangle *rect)
{
  GdkWindowObject         *private;
  GdkDrawableImplDirectFB *impl;
  
  g_return_if_fail (GDK_IS_WINDOW (window));
  g_return_if_fail (rect != NULL);
  
  if (GDK_WINDOW_DESTROYED (window))
    return;

  private = GDK_WINDOW_OBJECT (window);

  while (private->parent && ((GdkWindowObject*) private->parent)->parent)
    private = (GdkWindowObject*) private->parent;
  if (GDK_WINDOW_DESTROYED (window))
    return;

  impl = GDK_DRAWABLE_IMPL_DIRECTFB (private->impl);
  
  rect->x = impl->abs_x;
  rect->y = impl->abs_x;
  rect->width = impl->width;
  rect->height = impl->height;
}
